from typing import Callable, List, Union, overload

import reacton.ipyvuetify as v

import solara


@overload
def ConfirmationDialog(
    open: solara.Reactive[bool],
    *,
    on_close: Union[None, Callable[[], None]] = None,
    content: Union[str, solara.Element] = "",
    title: str = "Confirm action",
    ok: Union[str, solara.Element] = "OK",
    on_ok: Callable[[], None] = lambda: None,
    cancel: Union[str, solara.Element] = "Cancel",
    on_cancel: Callable[[], None] = lambda: None,
    children: List[solara.Element] = [],
    max_width: Union[int, str] = 500,
    persistent: bool = False,
):
    ...


# when open is a boolean, on_close should be given, otherwise a dialog can never be closed
# TODO: copy this pattern to many other components
@overload
def ConfirmationDialog(
    open: bool,
    *,
    on_close: Callable[[], None],
    content: Union[str, solara.Element] = "",
    title: str = "Confirm action",
    ok: Union[str, solara.Element] = "OK",
    on_ok: Callable[[], None] = lambda: None,
    cancel: Union[str, solara.Element] = "Cancel",
    on_cancel: Callable[[], None] = lambda: None,
    children: List[solara.Element] = [],
    max_width: Union[int, str] = 500,
    persistent: bool = False,
):
    ...


@solara.component
def ConfirmationDialog(
    open: Union[solara.Reactive[bool], bool],
    *,
    on_close: Union[None, Callable[[], None]] = None,
    content: Union[str, solara.Element] = "",
    title: str = "Confirm action",
    ok: Union[str, solara.Element] = "OK",
    on_ok: Callable[[], None] = lambda: None,
    cancel: Union[str, solara.Element] = "Cancel",
    on_cancel: Callable[[], None] = lambda: None,
    children: List[solara.Element] = [],
    max_width: Union[int, str] = 500,
    persistent: bool = False,
):
    """A dialog used to confirm a user action.

    (*Note: [This component is experimental and its API may change in the future](/docs/lab).*)

    By default, has a title, a text explaining the
    decision to be made, and two buttons "OK" and "Cancel".

    ## Basic examples

    ```solara
    import solara

    open_delete_confirmation = solara.reactive(False)

    def delete_user():
        # put your code to perform the action here
        print("User being deleted...")

    @solara.component
    def Page():
        solara.Button(label="Delete user", on_click=lambda: open_delete_confirmation.set(True))
        solara.lab.ConfirmationDialog(open_delete_confirmation, ok="Ok, Delete", on_ok=delete_user, content="Are you sure you want to delete this user?")
    ```

    ## Arguments

    * `open`: Indicates whether the dialog is being shown or not.
    * `on_open`: Callback to call when the dialog opens of closes.
    * `content`: Message that is displayed.
    * `title`: Title of the dialog.
    * `ok`: If a string, this text will be displayed on the confirmation button (default is "OK"). If a Button, it will be used instead of the default button.
    * `on_ok`: Callback to be called when the OK button is clicked.
    * `cancel`: If a string, this text will be displayed on the cancellation button (default is "Cancel"). If a Button, it will be used instead of the default
        button.
    * `on_cancel`: Callback to be called when the Cancel button is clicked. When persistent is False, clicking outside of the element or pressing esc key will
       also trigger cancel.
    * `children`: Additional components that will be shown under the dialog message, but before the buttons.
    * `max_width`: Maximum width of the dialog window.
    * `persistent`: When False (the default), clicking outside of the element or pressing esc key will trigger cancel.

    """

    def on_open(open_value):
        if not open_value:
            if on_close:
                on_close()

    open_reactive = solara.use_reactive(open, on_open)
    del open

    def close():
        open_reactive.set(False)

    user_on_click_ok = None
    user_on_click_cancel = None

    def perform_ok():
        if user_on_click_ok:
            user_on_click_ok()
        on_ok()
        close()

    def perform_cancel():
        if user_on_click_cancel:
            user_on_click_cancel()
        on_cancel()
        close()

    def on_v_model(value):
        if not value:
            on_cancel()
        open_reactive.set(value)

    with v.Dialog(
        v_model=open_reactive.value,
        on_v_model=on_v_model,
        persistent=persistent,
        max_width=max_width,
    ):
        with solara.v.Card():
            solara.v.CardTitle(children=[title])
            with solara.v.CardText(style_="min-height: 64px"):
                if isinstance(content, str):
                    solara.Text(content)
                else:
                    solara.display(content)
                if children:
                    solara.display(*children)
            with solara.v.CardActions(class_="justify-end"):
                if isinstance(cancel, str):
                    solara.Button(label=cancel, on_click=perform_cancel, text=True, classes=["grey--text", "text--darken-2"])
                else:
                    # the user may have passed in on_click already
                    user_on_click_cancel = cancel.kwargs.get("on_click")
                    # override or add our own on_click handler
                    cancel.kwargs = {**cancel.kwargs, "on_click": perform_cancel}
                    solara.display(cancel)

                # similar as cancel
                if isinstance(ok, str):
                    solara.Button(label=ok, on_click=perform_ok, dark=True, color="primary", elevation=0)
                else:
                    user_on_click_ok = ok.kwargs.get("on_click")
                    ok.kwargs = {**ok.kwargs, "on_click": perform_ok}
                    solara.display(ok)



import solara

from . import SharedComponent

github_url = solara.util.github_url(__file__)


@solara.component
def Sub():
    with solara.Card("Sub component", margin=0, classes=["my-2"]):
        solara.Markdown("This sub component is the best")
        with solara.Sidebar():
            with solara.Card("Sidebar of sub component", margin=0, elevation=0):
                solara.Markdown("*Markdown* **is** 👍")
            SharedComponent()


@solara.component
def Page():
    with solara.Sidebar():
        with solara.Card("Sidebar of page 1", margin=0, elevation=0):
            solara.Markdown("Hi there 👋!")
            solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)
    with solara.Card("Page 1"):
        Sub()
        solara.Markdown("Page 1 is the best")



"""# Altair

[Altair](https://altair-viz.github.io/index.html) is a declarative statistical visualization library for Python.

This example show how to use the [on_click handler](/api/altair) to display data for a specific day in the chart.

Based on [an Altair example](https://altair-viz.github.io/gallery/annual_weather_heatmap.html)

"""
import altair as alt
import pandas as pd
from vega_datasets import data

import solara

# title = "Altair visualization"
source = data.seattle_weather()

selected_datum = solara.reactive(None)


@solara.component
def Page():
    def on_click(datum):
        selected_datum.value = datum

    chart = (
        alt.Chart(source, title="Daily Max Temperatures (C) in Seattle, WA")
        .mark_rect()
        .encode(
            alt.X("date(date):O").title("Day").axis(format="%e", labelAngle=0),
            alt.Y("month(date):O").title("Month"),
            alt.Color("max(temp_max)").title(None),
            tooltip=[
                alt.Tooltip("monthdate(date)", title="Date"),
                alt.Tooltip("max(temp_max)", title="Max Temp"),
            ],
        )
        .configure_view(step=13, strokeWidth=0)
        .configure_axis(domain=False)
    )
    with solara.Card("Annual Weather Heatmap for Seattle, WA"):
        solara.AltairChart(chart, on_click=on_click)
        df = source
        if selected_datum.value:
            month_date = selected_datum.value["monthdate_date_end"]
            dt = pd.to_datetime(month_date)
            df = df[(df["date"].dt.month == dt.month) & (df["date"].dt.day == dt.day)]
            solara.Markdown(f"Day of year: {dt.month_name()} - {dt.day}")
            solara.Button("Clear selection", on_click=lambda: selected_datum.set(None))
            solara.display(df)

            with solara.Details("Click data"):
                solara.Markdown(
                    f"""
                Click data:

                ```
                {selected_datum.value}
                ```
                """
                )
        else:
            solara.Markdown("Click on the chart to see data for a specific day")



import traceback
from typing import Any, Dict

try:
    import comm
except ImportError:
    comm = None  # type: ignore

orphan_comm_stacks: Dict[Any, str] = {}


if comm is not None and comm.create_comm is comm._create_comm:
    # only when nobody else has monkey-patched comm.create_comm
    class DummyComm(comm.base_comm.BaseComm):  # type: ignore
        def publish_msg(self, msg_type, data=None, metadata=None, buffers=None, **keys):
            pass

    def create_dummy_comm(*args, **kwargs):
        comm = DummyComm(*args, **kwargs)
        stacktrace = "".join(traceback.format_stack())
        orphan_comm_stacks[comm] = stacktrace
        return comm

    comm.create_comm = create_dummy_comm
else:

    class DummyComm:  # type: ignore
        pass



import threading
import time
from pathlib import Path
from typing import Optional

from reacton import component
from reacton import ipywidgets as w
from reacton import render_fixed

import solara
from solara.datatypes import FileContentResult
from solara.hooks import use_download, use_fetch, use_json, use_thread

from .common import busy_wait_compare


def test_hook_thread():
    done = False

    @component
    def ThreadTest():
        nonlocal done
        a, _ = solara.use_state("a")
        b, _ = solara.use_state("b")
        counter, set_counter = solara.use_state(0)
        other, set_other = solara.use_state(0)
        c, _ = solara.use_state("c")
        if counter % 2 == 0:
            set_other(counter)

        def work(cancelled):
            for i in range(256):
                set_counter(i)
                time.sleep(0.001)

        use_thread(work, [])
        return w.Label(value=f"{a} {b} {c} {counter}")

    label, rc = render_fixed(ThreadTest(), handle_error=False)
    expected = "a b c 255"
    busy_wait_compare(lambda: label.value, expected)
    assert label.value == expected


def test_use_thread_keep_previous():
    def set_x(x):
        pass

    @component
    def Test():
        nonlocal set_x
        x, set_x = solara.use_state(2)

        def work():
            time.sleep(0.1)
            return x**2

        result: solara.Result[int] = use_thread(work, dependencies=[x])
        return w.Label(value=f"{result.value}")

    label, rc = render_fixed(Test(), handle_error=False)
    expected_2 = "4"
    busy_wait_compare(lambda: label.value, expected_2)
    assert label.value == expected_2

    expected_3 = "9"
    set_x(3)
    busy_wait_compare(lambda: label.value, expected_3)
    assert label.value == expected_3


def test_hook_iterator():
    event = threading.Event()
    result = None

    @solara.component
    def Test():
        nonlocal result

        def work():
            yield 1
            event.wait()
            yield 2

        result = use_thread(work)
        return w.Label(value="test")

    label, rc = solara.render_fixed(Test())
    assert result is not None
    assert isinstance(result, solara.Result)
    time.sleep(0.05)
    assert result.value == 1
    event.set()
    time.sleep(0.01)
    assert isinstance(result, solara.Result)
    assert result.value == 2


def test_use_thread_intrusive_cancel():
    result = None
    last_value = 0
    seconds = 4

    @solara.component
    def Test():
        nonlocal result
        nonlocal last_value

        def work():
            nonlocal last_value
            for i in range(100):
                last_value = i
                # if not cancelled, might take 4 seconds
                time.sleep(seconds / 100)
            return 2**42

        result = use_thread(work, dependencies=[])
        return w.Label(value="test")

    solara.render_fixed(Test())
    assert result is not None
    assert isinstance(result, solara.Result)
    result.cancel()
    while result.state == solara.ResultState.RUNNING:
        time.sleep(0.1)
    assert result.state == solara.ResultState.CANCELLED
    assert last_value != 99

    # also test retry
    seconds = 0.1
    result.retry()
    while result.state == solara.ResultState.CANCELLED:
        time.sleep(0.1)
    while result.state == solara.ResultState.RUNNING:
        time.sleep(0.1)
    assert last_value == 99


def test_hook_download(tmpdir):
    url = "https://raw.githubusercontent.com/widgetti/reacton/master/.gitignore"
    # content_length = 865

    path = tmpdir / "file.txt"

    @component
    def DownloadFile(url=url, expected_size=None):
        result = use_download(path, url, expected_size=expected_size)
        return w.Label(value=f"{result.progress} {result.state} {result.error}")

    label, rc = render_fixed(DownloadFile())
    assert label.value == "0 ResultState.RUNNING None"
    expected = "1.0 ResultState.FINISHED None"
    busy_wait_compare(lambda: label.value, expected)
    assert label.value == expected

    # if given, content_length, we should render with done immediately
    # there is not reliable way to test this, since it will still start the thread
    # label, rc = render_fixed(DownloadFile(expected_size=content_length), handle_error=False)
    # time.sleep(0.2)
    # assert label.value == expected

    label, rc = render_fixed(DownloadFile(url=url + ".404"))
    expected = "0 ResultState.ERROR HTTP Error 404: Not Found"
    busy_wait_compare(lambda: label.value, expected)
    assert label.value == expected


def test_hook_use_fetch():
    url = "https://raw.githubusercontent.com/widgetti/reacton/master/.gitignore"
    content_length = 888

    result = None

    @component
    def FetchFile(url=url, expected_size=None):
        nonlocal result
        result = use_fetch(url)
        data = result.value
        if result.error:
            raise result.error
        return w.Label(value=f"{len(data) if data else '-'}")

    label, rc = render_fixed(FetchFile(), handle_error=False)
    assert label.value == "-"
    expected = f"{content_length}"
    busy_wait_compare(lambda: label.value, expected)
    assert label.value == expected


def test_hook_use_json():
    url = "https://jherr-pokemon.s3.us-west-1.amazonaws.com/index.json"
    pokemons = 799

    result = None

    @component
    def FetchJson(url=url, expected_size=None):
        nonlocal result
        result = use_json(url)
        data = result.value
        return w.Label(value=f"{len(data) if data else '-'}")

    label, rc = render_fixed(FetchJson())
    assert label.value == "-"
    expected = f"{pokemons}"
    busy_wait_compare(lambda: label.value, expected)
    assert label.value == expected


def test_use_file_content(tmpdir: Path):
    path = tmpdir / "test.txt"
    path.write_text("Hi", "utf8")

    result: Optional[FileContentResult[bytes]] = None

    @solara.component
    def Test(path):
        nonlocal result
        result = solara.use_file_content(path)
        if result.value is not None:
            assert result.value == b"Hi"
            assert result.exists
        else:
            assert isinstance(result.error, FileNotFoundError)
            assert not result.exists
        # pickle.dumps(result)
        return w.Button()

    box, rc = solara.render(Test(path), handle_error=False)
    path_non_exist = tmpdir / "nonexist"
    rc.render(Test(path_non_exist))
    path_non_exist.write_text("Hi", "utf8")
    assert result is not None
    result.retry()
    assert result.value == b"Hi"


def test_use_state_or_update():
    values = []
    set_value = None

    @solara.component
    def Test(value):
        nonlocal set_value
        value, set_value = solara.use_state_or_update(value)
        values.append(value)
        return w.Button()

    container, rc = solara.render(Test(3))
    assert set_value is not None
    assert values == [3]
    set_value(5)
    assert values == [3, 5]
    rc.render(Test(9))
    assert values == [3, 5, 9, 9]  # we render twice, since we call set_state in or_update



from typing import Optional

import solara

from ... import data
from ...components.article import Overview


@solara.component
def Page(name: Optional[str] = None, page: int = 0, page_size=100):
    if name is None:
        with solara.Column() as main:
            solara.Title("Solara demo » Articles")
            Overview()
        return main
    if name not in data.articles:
        return solara.Error(f"No such article: {name!r}")
    article = data.articles[name]
    with solara.ColumnsResponsive(12) as main:
        solara.Title("Solara demo » Article » " + article.title)
        with solara.Link("/article"):
            solara.Text("« Back to overview")
        with solara.Card():
            pre = f"# {article.title}\n\n"
            solara.Markdown(pre + article.markdown)
    return main



import os
from pathlib import Path
from typing import Optional

import solara.settings

os.environ["MY_VALUE"] = "test"
os.environ["SOLARA_test_some_int"] = "42"
os.environ["SOLARA_test_some_bool"] = "true"
os.environ["SOLARA_test_value_no_field"] = "42"
os.environ["SOLARA_test_path"] = "/tmp/test"
os.environ["SOLARA_test_path_optional"] = "/tmp/test/optional"


class MySettings(solara.settings.BaseSettings):
    value: str = solara.settings.Field("default1", env="MY_VALUE")
    other_value: str = solara.settings.Field("default2", env="OTHER_VALUE")
    some_int: int = solara.settings.Field(128)
    some_other_int: int = solara.settings.Field(128)
    some_bool: bool = solara.settings.Field(False)
    some_other_bool: bool = solara.settings.Field(False)

    value_no_field: int = 50
    value_no_field_other: int = 50

    path: Path = Path("/tmp")
    path_optional: Optional[Path] = None

    class Config:
        env_prefix = "solara_test_"


def test_settings():
    settings = MySettings()
    assert settings.value == "test"
    assert settings.other_value == "default2"
    assert settings.some_int == 42
    assert settings.some_other_int == 128
    assert settings.some_bool is True
    assert settings.some_other_bool is False
    assert settings.value_no_field == 42
    assert settings.value_no_field_other == 50
    assert settings.path == Path("/tmp/test")
    assert settings.path_optional == Path("/tmp/test/optional")


def test_dict():
    settings = MySettings()
    test_value = "some-other-value"
    settings.value = test_value
    assert settings.dict()["value"] == test_value



from pathlib import Path

import playwright.sync_api
from flask import Flask

import solara
import solara.server.flask

try:
    from . import conftest
except ImportError:
    pass


HERE = Path(__file__).parent


@solara.component
def Page():
    solara.Markdown("Mounted in flask")


flask_app = Flask(__name__)
flask_app.register_blueprint(solara.server.flask.blueprint, url_prefix="/solara_mount")


@flask_app.route("/")
def hello_world():
    return "<p>Hello, World!</p>"


def test_flask_mount(page_session: playwright.sync_api.Page, solara_app, extra_include_path):
    port = conftest.TEST_PORT
    conftest.TEST_PORT += 1
    server = solara.server.flask.ServerFlask(port=port, flask_app=flask_app, url_prefix="/solara_mount")
    server.serve_threaded()
    server.wait_until_serving()
    with extra_include_path(HERE), solara_app("flask_test"):
        page_session.goto(f"{server.base_url}/solara_mount/")
        page_session.locator("text=Mounted in flask").wait_for()
        # assert page_session.text_content("p") == "Mounted in flask"
        # assert page_session.text_content("p") == "Mounted in flask"



from pathlib import Path

import solara
import solara.components.applayout
from solara.website.components.notebook import Notebook

HERE = Path(__file__).parent


@solara.component
def Page():
    # only execute once, other
    Notebook(HERE / "_data_science.ipynb")



from typing import Dict, List

import ipyvue
import traitlets

import solara


class SqlCodeWidget(ipyvue.VueTemplate):
    template_file = (__file__, "sql_code.vue")

    label = traitlets.Unicode("").tag(sync=True)
    query = traitlets.Unicode(allow_none=True, default_value=None).tag(sync=True)
    tables = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    height = traitlets.Unicode("180px").tag(sync=True)


@solara.component
def SqlCode(label="Query", query: str = None, tables: Dict[str, List[str]] = None, on_query=None, height="180px"):
    """SQL textfield input with auto complete and SQL syntax highlighting.

    To get auto complete for the column names, prefix it with the table name, i.e. "titanic.sur ctrl+space"

    ## Arguments

     * `label`: Label for the textfield.
     * `query`: SQL query.
     * `tables`: Dictionary with table names as keys and list of column names as values (used for auto complete).
     * `on_query`: Callback function that is called when the query is changed
     * `height`: Height of the textfield

    """
    return SqlCodeWidget.element(label=label, query=query, tables=tables, on_query=on_query, height=height)



import reacton.ipyvuetify as v
import solara
from solara.kitchensink import vue


@solara.component
def AppIcon(open=False, on_click=None):
    def click(*ignore):
        on_click()

    icon = v.AppBarNavIcon()
    v.use_event(icon, "click", click)
    return icon


@solara.component
def LayoutApp(children=[], left=None, right=None, open_left=False, open_right=False, title="Solara"):
    open_left, set_open_left = solara.use_state(open_left)
    open_right, set_open_right = solara.use_state(open_right)
    with v.Html(tag="div") as main:
        if left:
            with v.NavigationDrawer(absolute=True, right=False, width="min-content", v_model=open_left):
                AppIcon(open_left, on_click=lambda: set_open_left(not open_left))
                with v.Html(tag="div", children=[left]):
                    pass
        with v.Toolbar(dense=True, class_="elevation-3", dark=False):
            if left:
                AppIcon(open_left, on_click=lambda: set_open_left(not open_left))
            v.ToolbarTitle(children=[title])
            v.Spacer()
            if right:
                with v.Btn(icon=True) as btn:
                    vue.use_event(btn, "click", lambda *_ignore: set_open_right(not open_right))
                    v.Icon(children=["mdi-settings"])
        with v.Row():
            v.Col(cols=12, children=[*children])
        if right:
            with v.NavigationDrawer(absolute=True, right=True, width="min-content", v_model=open_right):
                with v.Btn(icon=True) as btn:
                    vue.use_event(btn, "click", lambda *_ignore: set_open_right(not open_right))
                    v.Icon(children=["mdi-settings"])
                with v.Html(tag="div", children=[right]):
                    pass
    return main



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Column(align="end", style={"height": "100%", "justify-content": "center"}):
            solara.Markdown(
                """
                    Team 5 built a tool for analyzing sports videos, which allows users to select or upload a video, and then analyse it using
                    point tracking, and object recognition. Because of the nature of the project, adding further analysis tools is very easy.

                    Note: The dashboard is no longer live on Ploomber, but you can see a preview of it in the video, or run it through
                    [GitHub](https://github.com/FelipeCabelloE/solarathon-team5/), using CodeSpaces.
                """
            )
        with solara.v.Html(tag="video", attributes={"controls": "controls", "autoplay": "autoplay"}, style_="width:100%;"):
            solara.v.Html(
                tag="source",
                attributes={"src": "https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_5/preview.webm", "type": "video/webm"},
            )



import os
from os.path import isfile, join
from pathlib import Path
from typing import Callable, List, Optional, Union, cast

import humanize
import ipyvuetify as vy
import traitlets

import solara
from solara.components import Div


def list_dir(path, filter: Callable[[Path], bool] = lambda x: True, directory_first: bool = False) -> List[dict]:
    def mk_item(n):
        full_path = join(path, n)
        is_file = isfile(full_path)
        return {"name": n, "is_file": is_file, "size": humanize.naturalsize(os.stat(full_path).st_size) if is_file else None}

    files = [mk_item(k) for k in os.listdir(path) if not k.startswith(".") if filter(Path(path) / k)]
    sorted_files = sorted(files, key=lambda item: (item["is_file"] == directory_first, item["name"].lower()))

    return sorted_files


class FileListWidget(vy.VuetifyTemplate):
    template_file = (__file__, "file_list_widget.vue")

    files = traitlets.List().tag(sync=True)
    clicked = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    double_clicked = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    scroll_pos = traitlets.Int(allow_none=True).tag(sync=True)

    def test_click(self, path: Union[Path, str], double_click=False):
        """Simulate a click or double click at the Python side"""
        matches = [k for k in self.files if k["name"] == str(path)]
        if len(matches) == 0:
            names = [k["name"] for k in self.files]
            raise NameError(f"Could not find {path}, possible filenames: {names}")
        item = matches[0]
        if double_click:
            self.double_clicked = item
        else:
            self.clicked = item

    def __contains__(self, name):
        """Test if filename/directory name is in the current directory."""
        return name in [k["name"] for k in self.files]


@solara.component
def FileBrowser(
    directory: Union[None, str, Path, solara.Reactive[Path]] = None,
    on_directory_change: Callable[[Path], None] = None,
    on_path_select: Callable[[Optional[Path]], None] = None,
    on_file_open: Callable[[Path], None] = None,
    filter: Callable[[Path], bool] = lambda x: True,
    directory_first: bool = False,
    on_file_name: Callable[[str], None] = None,
    start_directory=None,
    can_select=False,
):
    """File/directory browser at the server side.

    There are two modes possible

     * `can_select=False`
       * `on_file_open`: Triggered when **single** clicking a file or directory.
       * `on_path_select`: Never triggered
       * `on_directory_change`: Triggered when clicking a directory
     * `can_select=True`
       * `on_file_open`: Triggered when **double** clicking a file or directory.
       * `on_path_select`: Triggered when clicking a file or directory
       * `on_directory_change`: Triggered when double clicking a directory

    ## Arguments

     * `directory`: The directory to start in. If `None` the current working directory is used.
     * `on_directory_change`: Depends on mode, see above.
     * `on_path_select`: Depends on mode, see above.
     * `on_file_open`: Depends on mode, see above.
     * `filter`: A function that takes a `Path` and returns `True` if the file/directory should be shown.
     * `directory_first`: If `True` directories are shown before files. Default: `False`.
     * `on_file_name`: (deprecated) Use on_file_open instead.
     * `start_directory`: (deprecated) Use directory instead.
    """
    if start_directory is not None:
        directory = start_directory  # pragma: no cover
    if directory is None:
        directory = os.getcwd()  # pragma: no cover
    if isinstance(directory, str):
        directory = Path(directory)
    current_dir = solara.use_reactive(directory)
    selected, set_selected = solara.use_state(None)
    double_clicked, set_double_clicked = solara.use_state(None)
    warning, set_warning = solara.use_state(cast(Optional[str], None))
    scroll_pos_stack, set_scroll_pos_stack = solara.use_state(cast(List[int], []))
    scroll_pos, set_scroll_pos = solara.use_state(0)
    selected, set_selected = solara.use_state(None)

    def change_dir(new_dir: str):
        if os.access(new_dir, os.R_OK):
            current_dir.value = Path(new_dir)
            if on_directory_change:
                on_directory_change(Path(new_dir))
            set_warning(None)
            return True
        else:
            set_warning(f"[no read access to {new_dir}]")

    def on_item(item, double_click):
        if item is None:
            if can_select and on_path_select:
                on_path_select(None)
            return
        if item["name"] == "..":
            current_dir_str = str(current_dir.value)
            new_dir = current_dir_str[: current_dir_str.rfind(os.path.sep)]
            action_change_directory = (can_select and double_click) or (not can_select and not double_click)
            if action_change_directory and change_dir(new_dir):
                if scroll_pos_stack:
                    last_pos = scroll_pos_stack[-1]
                    set_scroll_pos_stack(scroll_pos_stack[:-1])
                    set_scroll_pos(last_pos)
                set_selected(None)
                set_double_clicked(None)
                if on_path_select and can_select:
                    on_path_select(None)
            if can_select and not double_click:
                if on_path_select:
                    on_path_select(Path(new_dir))
            return

        path = os.path.join(current_dir.value, item["name"])
        is_file = item["is_file"]
        if (can_select and double_click) or (not can_select and not double_click):
            if is_file:
                if on_file_open:
                    on_file_open(Path(path))
                if on_file_name is not None:
                    on_file_name(path)
            else:
                if change_dir(path):
                    set_scroll_pos_stack(scroll_pos_stack + [scroll_pos])
                    set_scroll_pos(0)
                set_selected(None)
            set_double_clicked(None)
            if on_path_select and can_select:
                on_path_select(None)
        elif can_select and not double_click:
            if on_path_select:
                on_path_select(Path(path))
        else:  # not can_select and double_click is ignored
            raise RuntimeError("Combination should not happen")  # pragma: no cover

    def on_click(item):
        set_selected(item)
        on_item(item, False)

    def on_double_click(item):
        set_double_clicked(item)
        if can_select:
            on_item(item, True)
        # otherwise we can ignore it, single click will handle it

    files = [{"name": "..", "is_file": False}] + list_dir(current_dir.value, filter=filter, directory_first=directory_first)
    with Div(class_="solara-file-browser") as main:
        Div(children=[str(current_dir.value)])
        FileListWidget.element(
            files=files,
            selected=selected,
            clicked=selected,
            on_clicked=on_click,
            double_clicked=double_clicked,
            on_double_clicked=on_double_click,
            scroll_pos=scroll_pos,
            on_scroll_pos=set_scroll_pos,
        ).key("FileList")
        if warning:
            Div(style_="font-weight: bold; color: red", children=[warning])

    return main



from typing import Optional

import ipyvuetify as vy
import reacton.core
import traitlets

import solara


class HeadTagWidget(vy.VuetifyTemplate):
    template_file = (__file__, "head_tag.vue")
    tagname = traitlets.Unicode().tag(sync=True)
    key = traitlets.Unicode(None, allow_none=True).tag(sync=True)
    attributes = traitlets.Dict().tag(sync=True)
    level = traitlets.Int().tag(sync=True)


@solara.component
def HeadTag(tagname: str, key=None, attributes: Optional[dict] = None):
    """Add a child element to head element, or replace a meta tag with the same tagname and key.

    This component should be used inside a [Head](/api/head) component, e.g.:

    ```python
    import solara

    @solara.component
    def Page():
        with solara.VBox() as main:
            MyAwesomeComponent()
            with solara.Head():
                solara.HeadTag(tagname="meta", attributes={"name": "description", "content": "My page description"})
        return main
    ```

    If multiple HeadTag components are used with the same key, the 'deepest' child will take precedence.

    ## Arguments

     * tagname: the tagname of the element (e.g. 'meta', 'link', 'script')
     * attributes: a dictionary of attributes to set on the element.
    """
    level = 0
    rc = reacton.core.get_render_context()
    context = rc.context
    while context and context.parent:
        level += 1
        context = context.parent
    attributes = attributes or {}
    return HeadTagWidget.element(tagname=tagname, key=key, attributes=attributes, level=level)



"""
# Menus

This page contains the various kinds of menu elements available to use in solara

# Menu
"""
import solara
from solara.website.utils import apidoc

from . import NoPage

title = "Menus"


__doc__ += apidoc(solara.lab.components.menu.Menu.f)  # type: ignore
__doc__ += "# ClickMenu"
__doc__ += apidoc(solara.lab.components.menu.ClickMenu.f)  # type: ignore
__doc__ += "# ContextMenu"
__doc__ += apidoc(solara.lab.components.menu.ContextMenu.f)  # type: ignore

Page = NoPage



"""
# GridFixed

Lays out children in a grid with a fixed number of columns.
"""

import solara

from .common import ColorCard

title = "GridFixed"


@solara.component
def Page():
    with solara.VBox() as main:
        colors = "green red orange brown yellow pink".split()
        with solara.GridFixed(columns=3):
            for color in colors:
                ColorCard(color, color)
    return main



import contextlib
import errno
import importlib
import inspect
import logging
import os
import sys
import threading
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Type

import solara.server.settings as settings

NO_WATCHDOG = False
try:
    from watchdog.events import FileSystemEventHandler
    from watchdog.observers import Observer
except ImportError:
    pass
    NO_WATCHDOG = True
logger = logging.getLogger("solara.server.reload")


class Watcher:
    def __init__(self, files, on_change: Callable[[str], None]):
        pass

    def close(self):
        pass

    def add_file(self, file):
        pass


class WatcherDummy:
    def __init__(self, files, on_change: Callable[[str], None]):
        pass

    def close(self):
        pass

    def add_file(self, file):
        pass


if NO_WATCHDOG:
    WatcherType: Type[Watcher] = WatcherDummy  # type: ignore

else:

    class WatcherWatchdog(FileSystemEventHandler):
        def __init__(self, files, on_change: Callable[[str], None]):
            self.on_change = on_change
            self.observers: List[Any] = []
            self.directories: Set[str] = set()
            self.files: List[str] = []
            self.mtimes: Dict[str, float] = dict()
            for file in files:
                self.add_file(file)

        def close(self):
            for o in self.observers:
                o.unschedule_all()
                o.stop()
            # this makes testing slow
            # for o in self.observers:
            #     o.join()

        def add_file(self, file):
            file = os.path.abspath(os.path.realpath(file))
            if file not in self.files:
                logger.debug("Watching file %s", file)
                if file in self.files:
                    raise RuntimeError(f"{file} was already added")
                self.files.append(file)
                self._watch_file(file)

        def _watch_file(self, file):
            self.mtimes[file] = os.path.getmtime(file)
            directory = os.path.realpath(os.path.dirname(file))
            self.watch_directory(directory)

        def watch_directory(self, directory):
            if directory not in self.directories:
                logger.debug("Watching directory %s", directory)
                observer = Observer()
                observer.schedule(self, directory, recursive=False)
                try:
                    observer.start()
                except OSError as e:
                    if e.errno == errno.ENOSPC:
                        # from rich import print()
                        from .utils import start_error

                        start_error(
                            "inotify watch limit reached", "Read https://github.com/widgetti/solara/#how-to-fix-inotify-watch-limit-reached", exception=e
                        )

                self.observers.append(observer)
                self.directories.add(directory)

        def on_modified(self, event):
            super().on_modified(event)
            logger.debug("Watch event: %s", event)
            if not event.is_directory:
                if event.src_path in self.files:
                    mtime_new = os.path.getmtime(event.src_path)
                    changed = mtime_new > self.mtimes[event.src_path]
                    self.mtimes[event.src_path] = mtime_new
                    if changed:
                        logger.debug("File modified: %s", event.src_path)
                        try:
                            self.on_change(event.src_path)
                        except:  # noqa
                            # we are in the watchdog thread here, all we can do is report
                            # and continue running (otherwise reload stops working)
                            logger.exception("Error while executing on change handler")
                    else:
                        logger.debug("File reported modified, but mtime did not change: %s", event.src_path)
                else:
                    logger.debug("Ignore file modification: %s", event.src_path)

    WatcherType = WatcherWatchdog  # type: ignore


class Reloader:
    def __init__(self, on_change: Optional[Callable[[str], None]] = None) -> None:
        self.watched_modules: Set[str] = set()
        self._first = True
        self.on_change: Optional[Callable[[str], None]] = on_change
        WatcherCls = WatcherType if settings.main.mode == "development" else WatcherDummy
        self.watcher = WatcherCls([], self._on_change)
        self.requires_reload = False
        self.ignore_modules: Set[str] = set()
        self.reload_event_next = threading.Event()
        # should be set at app.directory
        self.root_path: Optional[Path] = None
        # maybe we want this mode enabled in the future via configuration
        # this is useful if you have some packages installed in editable mode
        # and you want to quick reload. However this does not always work:
        #  * https://github.com/widgetti/solara/issues/177
        #  * https://github.com/widgetti/solara/issues/148
        self.aggresive_reload = False

    def start(self):
        if self._first:
            # although we try to keep track of modules loaded (which we will watch)
            # it can happen that an import is done at runtime, that we miss (could be in a thread)
            # so we always reload all modules except the ignore_modules
            self.ignore_modules = set(sys.modules)
            logger.debug("Ignoring reloading modules: %s", self.ignore_modules)
            self._first = False

    def _on_change(self, name):
        # used for testing
        self.reload_event_next.set()
        # flag that we need to reload all modules next time
        self.requires_reload = True
        # and forward callback
        if self.on_change:
            self.on_change(name)

    def close(self):
        self.watcher.close()

    def get_reload_module_names(self):
        if self.aggresive_reload:
            # not sure why, but if we reload pandas, the integration/reload_test.py fails
            return {
                k for k in set(sys.modules) - set(self.ignore_modules) if not (k.startswith("solara.server") or k.startswith("anyio") or k.startswith("pandas"))
            }
        else:
            reload = []
            for name in sorted(sys.modules):
                mod = sys.modules[name]
                if name.startswith("solara.server"):
                    continue  # this will break everything
                if name in self.ignore_modules:
                    continue  # nothing we imported from solara itself like starlette etc
                # we only reload modules that are in the root path
                if (getattr(mod, "__file__", None) or "").startswith(str(self.root_path)):
                    reload.append(name)
            return reload

    def reload(self):
        # before we did this:
        # # don't reload modules like solara.server and react
        # # that may cause issues (like 2 Element classes existing)
        reload_modules = self.get_reload_module_names()
        # which picks up import that are done in threads etc, but it will also reload starlette, httptools etc
        # which causes issues with exceptions and isinstance checks.
        # reload_modules = self.watched_modules
        logger.info("Reloading modules... %s", reload_modules)
        # not sure if this is needed
        importlib.invalidate_caches()
        for mod_name in sorted(reload_modules):
            # don't reload modules like solara.server and react
            # that may cause issues (like 2 Element classes existing)
            logger.debug("Reloading module %s", mod_name)
            sys.modules.pop(mod_name, None)
        # if all successful...
        self.requires_reload = False

    @contextlib.contextmanager
    def watch(self):
        """Use this context manager to execute code so that we can track loaded modules and reload them when needed"""
        if self.requires_reload:
            self.reload()

        modules_before = set(sys.modules)
        try:
            yield
        finally:
            modules_after = set(sys.modules)
            # these are imported during the 'yield'
            modules_added = modules_after - modules_before
            modules_to_consider = modules_added - set(self.watched_modules)
            # TODO: libraries that solara uses need special care
            # modules_always = {k for k in sys.modules if k.startswith("reacton")}
            # modules_to_consider = modules_to_consider | modules_always
            modules_watching = set()
            if modules_to_consider:
                logger.debug("Found modules %s", modules_to_consider)
            for modname in modules_to_consider:
                module = sys.modules[modname]
                path = None
                try:
                    path = inspect.getfile(module)
                except Exception:
                    pass
                if path and Path(path).exists():
                    if not path.startswith(sys.prefix):
                        self.watcher.add_file(path)
                        self.watched_modules.add(modname)
                        modules_watching.add(modname)
            if modules_watching:
                logger.debug("Watching modules: %s for reload", modules_watching)


# there is only a reloader, and there should be only 1 app
# that connect to the on_change
reloader = Reloader()



"""# Custom state storage

Solara makes it easy to store state/data on the server side, scoped to a kernel, using [reactive variables](/api/reactive).

However, sometimes you want to store state yourself in an external system (i.e. not Solara), and for this you can use the
[get_kernel_id()](/api/get_kernel_id) function to get a unique id for each kernel.

If you want to store state/data scoped to a browser session, you can use the [get_session_id()](/api/get_session_id)
function to get a unique id tied to the users browser. This can be used to store state that outlives a page refresh.

In case you want to store state/data scoped to a user, you can use a similar strategy, but use a unique identifier based on the user,
instead of the session id. You can take a look at [Our oauth example](examples/general/login_oauth) or
[the authorization example](/examples/apps/authorization) for inspiration.

"""
from typing import Dict

import solara
import solara.lab

# used only to force updating of the page
force_update_counter = solara.reactive(0)

# Kernel storage is scoped to the kernel, and will be cleared when the kernel is stopped.
kernel_storage: Dict[str, str] = {}


def store_in_kernel_storage(value):
    kernel_storage[solara.get_kernel_id()] = value
    force_update_counter.value += 1


@solara.lab.on_kernel_start
def initialize_kernel_storage():
    # when a kernel gets started, we initialize the dict entry
    kernel_storage[solara.get_kernel_id()] = "This does not"

    def cleanup():
        # when a kernel gets stopped, we remove the dict entry
        del kernel_storage[solara.get_kernel_id()]

    # cleaning up kernel storage, we prevent memory leaks
    return cleanup


# session storage has no lifecycle management, and will only be cleared when the server is restarted.
session_storage: Dict[str, str] = {}


def store_in_session_storage(value):
    session_storage[solara.get_session_id()] = value
    force_update_counter.value += 1


@solara.component
def Page():
    solara.InputText(
        "Stored under the kernel id key",
        value=kernel_storage[solara.get_kernel_id()],
        on_value=store_in_kernel_storage,
        continuous_update=True,
    )

    solara.InputText(
        "Stored under the session id key",
        value=session_storage.get(solara.get_session_id(), "This outlives a page refresh"),
        on_value=store_in_session_storage,
        continuous_update=True,
    )



from typing import Callable, Optional, TypeVar, Union

import solara

T = TypeVar("T")


def use_reactive(
    value: Union[T, solara.Reactive[T]],
    on_change: Optional[Callable[[T], None]] = None,
) -> solara.Reactive[T]:
    """Creates a reactive variable with the a local component scope.

    It is a useful alternative to `use_state` when you want to use a
    reactive variable for the component state.
    See also [our documentation on state management](/docs/fundamentals/state-management).

    If the variable passed is a reactive variable, it will be returned instead and no
    new reactive variable will be created. This is useful for implementing component
    that accept either a reactive variable or a normal value along with an optional `on_change`
    callback.

    Note that that on each call, if the value changes, the reactive variable will be updated.
    For objects that do no implement equality comparison, the will lead to an infinite loop.

    In that case, combine the `use_reactive` with `use_memo` to never trigger an update from
    the render function.

    ```python
    @solara.component
    def Page():
        data = solara.use_reactive(solara.use_memo(lambda: MyDataObject()))
        ...
    ```

    ## Arguments:

     * value (Union[T, solara.Reactive[T]]): The value of the
            reactive variable. If a reactive variable is provided, it will be
            used directly. Otherwise, a new reactive variable will be created
            with the provided initial value. If the argument passed changes
            the reactive variable will be updated.

     * on_change (Optional[Callable[[T], None]]): An optional callback function
            that will be called when the reactive variable's value changes.

    Returns:
        solara.Reactive[T]: A reactive variable with the specified initial value
            or the provided reactive variable.

    ## Examples

    ### Replacement for use_state
    ```solara
    import solara

    @solara.component
    def ReusableComponent():
        color = solara.use_reactive("red")  # another possibility
        solara.Select(label="Color",values=["red", "green", "blue", "orange"],
                    value=color)
        solara.Markdown("### Solara is awesome", style={"color": color.value})

    @solara.component
    def Page():
        # this component is used twice, but each instance has its own state
        ReusableComponent()
        ReusableComponent()

    ```

    ### Flexible arguments

    The `MyComponent` component can be passed a reactive variable or a normal
    Python variable and a `on_value` callback.

    ```python
    import solara
    from typing import Union, Optional, Callable

    @solara.component
    def MyComponent(value: Union[T, solara.Reactive[T]],
                    on_value: Optional[Callable[[T], None]] = None,
        ):
        reactive_value = solara.use_reactive(value, on_value)
        # Use the `reactive_value` in the component
    ```
    """

    try:
        on_change_ref = solara.use_ref(on_change)
    except RuntimeError as e:
        raise RuntimeError(
            "use_reactive must be called from a component function, inside the render function.\n"
            "Do not call it top level, use [solara.reactive()](https://solara.dev/api/reactive) instead."
        ) from e
    on_change_ref.current = on_change

    def create():
        if not isinstance(value, solara.Reactive):
            return solara.reactive(value)

    reactive_value = solara.use_memo(create, dependencies=[])
    if isinstance(value, solara.Reactive):
        reactive_value = value
    assert reactive_value is not None
    updating = solara.use_ref(False)

    def forward_on_change():
        def forward(value):
            if on_change_ref.current and not updating.current:
                on_change_ref.current(value)

        return reactive_value.subscribe(forward)

    def update():
        updating.current = True
        try:
            if not isinstance(value, solara.Reactive):
                reactive_value.value = value
        finally:
            updating.current = False

    solara.use_memo(update, [value])
    # if value is a reactive variable, and it changes, we need to subscribe to the latest
    # reactive variable, otherwise we only link to it once
    solara.use_effect(forward_on_change, [value] if isinstance(value, solara.Reactive) else [])

    return reactive_value



import inspect
from typing import Any, Callable, Dict, Type

import ipyvue as vue
import ipyvuetify as v
import ipywidgets as widgets
import traitlets
import typing_extensions

import solara

P = typing_extensions.ParamSpec("P")


def _widget_from_signature(classname, base_class: Type[widgets.Widget], func: Callable[..., None], event_prefix: str) -> Type[widgets.Widget]:
    classprops: Dict[str, Any] = {}

    parameters = inspect.signature(func).parameters
    for name, param in parameters.items():
        if name.startswith("event_"):
            event_name = name[6:]

            def event_handler(self, data, buffers=None, event_name=event_name):
                callback = self._event_callbacks.get(event_name)
                if callback:
                    if buffers:
                        callback(data, buffers)
                    else:
                        callback(data)

            classprops[f"vue_{event_name}"] = event_handler
        if name.startswith("on_") and name[3:] in parameters:
            # callback, will be handled by reacton
            continue
        if param.default == inspect.Parameter.empty:
            trait = traitlets.Any()
        else:
            trait = traitlets.Any(default_value=param.default)
        classprops[name] = trait.tag(sync=True, **widgets.widget_serialization)
    # maps event_foo to a callable
    classprops["_event_callbacks"] = traitlets.Dict(default_value={})

    widget_class = type(classname, (base_class,), classprops)
    return widget_class


def _widget_vue(vue_path: str, vuetify=True) -> Callable[[Callable[P, None]], Type[v.VuetifyTemplate]]:
    def decorator(func: Callable[P, None]):
        class VuetifyWidgetSolara(v.VuetifyTemplate):
            template_file = (inspect.getfile(func), vue_path)

        class VueWidgetSolara(vue.VueTemplate):
            template_file = (inspect.getfile(func), vue_path)

        base_class = VuetifyWidgetSolara if vuetify else VueWidgetSolara
        widget_class = _widget_from_signature("VueWidgetSolaraSub", base_class, func, "vue_")

        return widget_class

    return decorator


def component_vue(vue_path: str, vuetify=True) -> Callable[[Callable[P, None]], Callable[P, solara.Element]]:
    """Decorator to create a component backed by a Vue template.

    Although many components can be made from the Python side, sometimes it is easier to write components using Vue directly.
    It can also be beneficial for performance, since instead of creating many widgets from the Python side we only send data to
    the frontend. If event handling is also done on the frontend, this reduces latency and makes you app feel much smoother.


    All arguments of the function are exposed as Vue properties. Argument pairs of the form `foo`, and `on_foo`
    are assumed by refer to the same vue property, with `on_foo` being the event handler when `foo` changes from
    the vue template.

    Arguments or the form `event_foo` should be callbacks that can be called from the vue template. They are
    available as the function `foo` in the vue template.

    [See the vue v2 api](https://v2.vuejs.org/v2/api/) for more information on how to use Vue, like `watch`,
    `methods` and lifecycle hooks such as `mounted` and `destroyed`.

    See the [Vue component example](/examples/general/vue_component) for an example of how to use this decorator.

    ## Arguments

     * `vue_path`: The path to the Vue template file.
     * `vuetify`: Whether the Vue template uses Vuetify components.

    """

    def decorator(func: Callable[P, None]):
        VueWidgetSolaraSub = _widget_vue(vue_path, vuetify=vuetify)(func)

        def wrapper(*args, **kwargs):
            event_callbacks = {}
            kwargs = kwargs.copy()
            # take out all events named like event_foo and put them in a separate dict
            for name in list(kwargs):
                if name.startswith("event_"):
                    event_callbacks[name[6:]] = kwargs.pop(name)
            if event_callbacks:
                kwargs["_event_callbacks"] = event_callbacks
            return VueWidgetSolaraSub.element(*args, **kwargs)  # type: ignore

        return wrapper

    return decorator


_component_vue = component_vue



import asyncio
import json
import logging
import math
import os
import sys
import threading
import typing
from typing import Any, Dict, List, Optional, Union, cast
from uuid import uuid4

import anyio
import starlette.websockets
import uvicorn.server
import websockets.legacy.http

try:
    import solara_enterprise

    del solara_enterprise
    has_solara_enterprise = True
except ImportError:
    has_solara_enterprise = False
if has_solara_enterprise and sys.version_info[:2] > (3, 6):
    has_auth_support = True
    from solara_enterprise.auth.middleware import MutateDetectSessionMiddleware
    from solara_enterprise.auth.starlette import (
        AuthBackend,
        authorize,
        get_user,
        login,
        logout,
    )
else:
    has_auth_support = False

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import HTTPConnection, Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles
from starlette.types import Receive, Scope, Send

import solara
import solara.settings
from solara.server.threaded import ServerBase

from . import app as appmod
from . import kernel_context, server, settings, telemetry, websocket
from .cdn_helper import cdn_url_path, get_path

os.environ["SERVER_SOFTWARE"] = "solara/" + str(solara.__version__)
limiter: Optional[anyio.CapacityLimiter] = None
lock = threading.Lock()


def _ensure_limiter():
    # in older anyios (<4) the limiter can only be created in an async context
    # so we call this in a starlette handler
    global limiter
    if limiter is None:
        with lock:
            if limiter is None:
                limiter = anyio.CapacityLimiter(settings.kernel.max_count if settings.kernel.max_count is not None else math.inf)


logger = logging.getLogger("solara.server.fastapi")
# if we add these to the router, the server_test does not run (404's)
prefix = ""

# The limit for starlette's http traffic should come from h11's DEFAULT_MAX_INCOMPLETE_EVENT_SIZE=16kb
# In practice, testing with 132kb cookies (server_test.py:test_large_cookie) seems to work fine.
# For the websocket, the limit is set to 4kb till 10.4, see
#  * https://github.com/aaugustin/websockets/blob/10.4/src/websockets/legacy/http.py#L14
# Later releases should set this to 8kb. See
#  * https://github.com/aaugustin/websockets/commit/8ce4739b7efed3ac78b287da7fb5e537f78e72aa
#  * https://github.com/aaugustin/websockets/issues/743
# Since starlette seems to accept really large values for http, lets do the same for websockets
# An arbitrarily large value we settled on for now is 32kb
# If we don't do this, users with many cookies will fail to get a websocket connection.
websockets.legacy.http.MAX_LINE = 1024 * 32


class WebsocketDebugInfo:
    lock = threading.Lock()
    attempts = 0
    connecting = 0
    open = 0
    closed = 0


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: starlette.websockets.WebSocket

    def __init__(self, ws: starlette.websockets.WebSocket, portal: anyio.from_thread.BlockingPortal) -> None:
        self.ws = ws
        self.portal = portal
        self.to_send: List[Union[str, bytes]] = []
        if settings.main.experimental_performance:
            self.task = asyncio.ensure_future(self.process_messages_task())

    async def process_messages_task(self):
        while True:
            await asyncio.sleep(0.05)
            while len(self.to_send) > 0:
                first = self.to_send.pop(0)
                if isinstance(first, bytes):
                    await self.ws.send_bytes(first)
                else:
                    await self.ws.send_text(first)

    def close(self):
        self.portal.call(self.ws.close)  # type: ignore

    def send_text(self, data: str) -> None:
        if settings.main.experimental_performance:
            self.to_send.append(data)
        else:
            self.portal.call(self.ws.send_bytes, data)  # type: ignore

    def send_bytes(self, data: bytes) -> None:
        if settings.main.experimental_performance:
            self.to_send.append(data)
        else:
            self.portal.call(self.ws.send_bytes, data)  # type: ignore

    async def receive(self):
        if hasattr(self.portal, "start_task_soon"):
            # version 3+
            fut = self.portal.start_task_soon(self.ws.receive)  # type: ignore
        else:
            fut = self.portal.spawn_task(self.ws.receive)  # type: ignore

        message = await asyncio.wrap_future(fut)
        if "text" in message:
            return message["text"]
        elif "bytes" in message:
            return message["bytes"]
        elif message.get("type") == "websocket.disconnect":
            raise websocket.WebSocketDisconnect()
        else:
            raise RuntimeError(f"Unknown message type {message}")


class ServerStarlette(ServerBase):
    server: uvicorn.server.Server
    name = "starlette"

    def __init__(self, port: int, host: str = "localhost", starlette_app=None, **kwargs):
        super().__init__(port, host, **kwargs)
        self.app = starlette_app or app

    def has_started(self):
        return self.server.started

    def signal_stop(self):
        self.server.should_exit = True
        # this cause uvicorn to not wait for background tasks, e.g.:
        # <Task pending name='Task-55'
        #  coro=<WebSocketProtocol.run_asgi() running at
        #  /.../uvicorn/protocols/websockets/websockets_impl.py:184>
        # wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x16896aa00>()]>
        # cb=[WebSocketProtocol.on_task_complete()]>
        self.server.force_exit = True
        self.server.lifespan.should_exit = True

    def serve(self):
        from uvicorn.config import Config
        from uvicorn.server import Server

        if sys.version_info[:2] < (3, 7):
            # make python 3.6 work
            import asyncio

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # uvloop will trigger a: RuntimeError: There is no current event loop in thread 'fastapi-thread'
        config = Config(self.app, host=self.host, port=self.port, **self.kwargs, access_log=False, loop="asyncio")
        self.server = Server(config=config)
        self.started.set()
        self.server.run()


async def kernels(id):
    return JSONResponse({"name": "lala", "id": "dsa"})


async def kernel_connection(ws: starlette.websockets.WebSocket):
    _ensure_limiter()
    try:
        with WebsocketDebugInfo.lock:
            WebsocketDebugInfo.attempts += 1
            WebsocketDebugInfo.connecting += 1
        await _kernel_connection(ws)
    finally:
        with WebsocketDebugInfo.lock:
            WebsocketDebugInfo.closed += 1
            WebsocketDebugInfo.open -= 1


async def _kernel_connection(ws: starlette.websockets.WebSocket):
    session_id = ws.cookies.get(server.COOKIE_KEY_SESSION_ID)

    if settings.oauth.private and not has_auth_support:
        raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
    if has_auth_support and "session" in ws.scope:
        user = get_user(ws)
        if user is None and settings.oauth.private:
            await ws.accept()
            logger.error("app is private, requires login")
            await ws.close(code=1008, reason="app is private, requires login")
            return
    else:
        user = None

    if not session_id:
        logger.warning("no session cookie")
        session_id = "session-id-cookie-unavailable:" + str(uuid4())
    # we use the jupyter session_id query parameter as the key/id
    # for a page scope.
    page_id = ws.query_params["session_id"]
    if not page_id:
        logger.error("no page_id")
    kernel_id = ws.path_params["kernel_id"]
    if not kernel_id:
        logger.error("no kernel_id")
        await ws.close()
        return
    logger.info("Solara kernel requested for session_id=%s kernel_id=%s", session_id, kernel_id)
    await ws.accept()
    with WebsocketDebugInfo.lock:
        WebsocketDebugInfo.connecting -= 1
        WebsocketDebugInfo.open += 1

    def websocket_thread_runner(ws: starlette.websockets.WebSocket, portal: anyio.from_thread.BlockingPortal):
        async def run():
            try:
                assert session_id is not None
                assert kernel_id is not None
                telemetry.connection_open(session_id)
                headers_dict: Dict[str, List[str]] = {}
                for k, v in ws.headers.items():
                    if k not in headers_dict.keys():
                        headers_dict[k] = [v]
                    else:
                        headers_dict[k].append(v)
                await server.app_loop(ws_wrapper, ws.cookies, headers_dict, session_id, kernel_id, page_id, user)
            except:  # noqa
                await portal.stop(cancel_remaining=True)
                raise
            finally:
                telemetry.connection_close(session_id)

        # sometimes throws: RuntimeError: Already running asyncio in this thread
        anyio.run(run)  # type: ignore

    # this portal allows us to sync call the websocket calls from this current event loop we are in
    # each websocket however, is handled from a separate thread
    try:
        async with anyio.from_thread.BlockingPortal() as portal:
            ws_wrapper = WebsocketWrapper(ws, portal)
            thread_return = anyio.to_thread.run_sync(websocket_thread_runner, ws, portal, limiter=limiter)  # type: ignore
            await thread_return
    finally:
        if settings.main.experimental_performance:
            try:
                ws_wrapper.task.cancel()
            except:  # noqa
                logger.exception("error cancelling websocket task")
        try:
            await ws.close()
        except:  # noqa
            pass


def close(request: Request):
    kernel_id = request.path_params["kernel_id"]
    page_id = request.query_params["session_id"]
    if kernel_id in kernel_context.contexts:
        context = kernel_context.contexts[kernel_id]
        context.page_close(page_id)
    response = HTMLResponse(content="", status_code=200)
    return response


async def root(request: Request, fullpath: str = ""):
    if settings.oauth.private and not has_auth_support:
        raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
    root_path = settings.main.root_path or ""
    if not settings.main.base_url:
        settings.main.base_url = str(request.base_url)
    # if not explicltly set,
    if settings.main.root_path is None:
        # use the default root path from the app, which seems to also include the path
        # if we are mounted under a path
        scope = request.scope
        root_path = scope.get("route_root_path", scope.get("root_path", ""))
        logger.debug("root_path: %s", root_path)
        # or use the script-name header, for instance when running under a reverse proxy
        script_name = request.headers.get("script-name")
        if script_name:
            logger.debug("override root_path using script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        script_name = request.headers.get("x-script-name")
        if script_name:
            logger.debug("override root_path using x-script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        settings.main.root_path = root_path

    request_path = request.url.path
    if request_path.startswith(root_path):
        request_path = request_path[len(root_path) :]
    content = server.read_root(request_path, root_path)
    if content is None:
        if settings.oauth.private and not request.user.is_authenticated:
            raise HTTPException(status_code=401, detail="Unauthorized")
        return HTMLResponse(content="Page not found by Solara router", status_code=404)

    if settings.oauth.private and not request.user.is_authenticated:
        from solara_enterprise.auth.starlette import login

        return await login(request)

    response = HTMLResponse(content=content)
    session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID) or str(uuid4())
    samesite = "lax"
    secure = False
    # we want samesite, so we can set a cookie when embedded in an iframe, such as on huggingface
    # however, samesite=none requires Secure https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite
    # when hosted on the localhost domain we can always set the Secure flag
    # to allow samesite https://developer.mozilla.org/en-US/docs/Web/HTTP/Cookies#restrict_access_to_cookies
    if request.headers.get("x-forwarded-proto", "http") == "https" or request.base_url.hostname == "localhost":
        samesite = "none"
        secure = True
    response.set_cookie(
        server.COOKIE_KEY_SESSION_ID, value=session_id, expires="Fri, 01 Jan 2038 00:00:00 GMT", samesite=samesite, secure=secure  # type: ignore
    )  # type: ignore
    return response


class StaticFilesOptionalAuth(StaticFiles):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        conn = HTTPConnection(scope)
        if settings.oauth.private and not has_auth_support:
            raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
        if has_auth_support and settings.oauth.private and not conn.user.is_authenticated:
            raise HTTPException(status_code=401, detail="Unauthorized")
        await super().__call__(scope, receive, send)


class StaticNbFiles(StaticFilesOptionalAuth):
    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        return cast(List[Union[str, "os.PathLike[str]"]], server.nbextensions_directories)

    # follow symlinks
    # from https://github.com/encode/starlette/pull/1377/files
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        for directory in self.all_directories:
            original_path = os.path.join(directory, path)
            full_path = os.path.realpath(original_path)
            directory = os.path.realpath(directory)
            try:
                return full_path, os.stat(full_path)
            except (FileNotFoundError, NotADirectoryError):
                continue
        return "", None


class StaticPublic(StaticFilesOptionalAuth):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        return cast(List[Union[str, "os.PathLike[str]"]], [app.directory.parent / "public" for app in appmod.apps.values()])


class StaticAssets(StaticFilesOptionalAuth):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        overrides = [app.directory.parent / "assets" for app in appmod.apps.values()]
        default = server.solara_static.parent / "assets"
        return cast(List[Union[str, "os.PathLike[str]"]], [*overrides, default])


class StaticCdn(StaticFilesOptionalAuth):
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        try:
            full_path = str(get_path(settings.assets.proxy_cache_dir, path))
        except Exception:
            return "", None

        return full_path, os.stat(full_path)


def on_startup():
    # TODO: configure and set max number of threads
    # see https://github.com/encode/starlette/issues/1724
    telemetry.server_start()


def on_shutdown():
    telemetry.server_stop()


def readyz(request: Request):
    json, status = server.readyz()
    return JSONResponse(json, status_code=status)


async def resourcez(request: Request):
    _ensure_limiter()
    assert limiter is not None
    data: Dict[str, Any] = {}
    verbose = request.query_params.get("verbose", None) is not None
    data["websockets"] = {
        "attempts": WebsocketDebugInfo.attempts,
        "connecting": WebsocketDebugInfo.connecting,
        "open": WebsocketDebugInfo.open,
        "closed": WebsocketDebugInfo.closed,
    }
    from . import patch

    data["threads"] = {
        "created": patch.ThreadDebugInfo.created,
        "running": patch.ThreadDebugInfo.running,
        "stopped": patch.ThreadDebugInfo.stopped,
        "active": threading.active_count(),
    }
    contexts = list(kernel_context.contexts.values())
    data["kernels"] = {
        "total": len(contexts),
        "has_connected": len([k for k in contexts if kernel_context.PageStatus.CONNECTED in k.page_status.values()]),
        "has_disconnected": len([k for k in contexts if kernel_context.PageStatus.DISCONNECTED in k.page_status.values()]),
        "has_closed": len([k for k in contexts if kernel_context.PageStatus.CLOSED in k.page_status.values()]),
        "limiter": {
            "total_tokens": limiter.total_tokens,
            "borrowed_tokens": limiter.borrowed_tokens,
            "available_tokens": limiter.available_tokens,
        },
    }
    default_limiter = anyio.to_thread.current_default_thread_limiter()
    data["anyio.to_thread.limiter"] = {
        "total_tokens": default_limiter.total_tokens,
        "borrowed_tokens": default_limiter.borrowed_tokens,
        "available_tokens": default_limiter.available_tokens,
    }
    if verbose:
        try:
            import psutil

            def expand(named_tuple):
                return {key: getattr(named_tuple, key) for key in named_tuple._fields}

            data["cpu"] = {}
            try:
                data["cpu"]["percent"] = psutil.cpu_percent()
            except Exception as e:
                data["cpu"]["percent"] = str(e)
            try:
                data["cpu"]["count"] = psutil.cpu_count()
            except Exception as e:
                data["cpu"]["count"] = str(e)
            try:
                data["cpu"]["times"] = expand(psutil.cpu_times())
                data["cpu"]["times"]["per_cpu"] = [expand(x) for x in psutil.cpu_times(percpu=True)]
            except Exception as e:
                data["cpu"]["times"] = str(e)
            try:
                data["cpu"]["times_percent"] = expand(psutil.cpu_times_percent())
                data["cpu"]["times_percent"]["per_cpu"] = [expand(x) for x in psutil.cpu_times_percent(percpu=True)]
            except Exception as e:
                data["cpu"]["times_percent"] = str(e)
            try:
                memory = psutil.virtual_memory()
            except Exception as e:
                data["memory"] = str(e)
            else:
                data["memory"] = {
                    "bytes": expand(memory),
                    "GB": {key: getattr(memory, key) / 1024**3 for key in memory._fields},
                }

        except ModuleNotFoundError:
            pass

    json_string = json.dumps(data, indent=2)
    return Response(content=json_string, media_type="application/json")


middleware = [
    Middleware(GZipMiddleware, minimum_size=1000),
]

if has_auth_support:
    middleware = [
        *middleware,
        Middleware(
            MutateDetectSessionMiddleware,
            secret_key=settings.session.secret_key,  # type: ignore
            session_cookie="solara-session",  # type: ignore
            https_only=settings.session.https_only,  # type: ignore
            same_site=settings.session.same_site,  # type: ignore
        ),
        Middleware(AuthenticationMiddleware, backend=AuthBackend()),
    ]

routes_auth = []
if has_auth_support:
    routes_auth = [
        Route("/_solara/auth/authorize", endpoint=authorize),  #
        Route("/_solara/auth/logout", endpoint=logout),
        Route("/_solara/auth/login", endpoint=login),
    ]
routes = [
    Route("/readyz", endpoint=readyz),
    Route("/resourcez", endpoint=resourcez),
    *routes_auth,
    Route("/jupyter/api/kernels/{id}", endpoint=kernels),
    WebSocketRoute("/jupyter/api/kernels/{kernel_id}/{name}", endpoint=kernel_connection),
    Route("/", endpoint=root),
    Route("/{fullpath}", endpoint=root),
    Route("/_solara/api/close/{kernel_id}", endpoint=close, methods=["POST"]),
    # only enable when the proxy is turned on, otherwise if the directory does not exists we will get an exception
    *([Mount(f"/{cdn_url_path}", app=StaticCdn(directory=settings.assets.proxy_cache_dir))] if solara.settings.assets.proxy else []),
    Mount(f"{prefix}/static/public", app=StaticPublic()),
    Mount(f"{prefix}/static/assets", app=StaticAssets()),
    Mount(f"{prefix}/static/nbextensions", app=StaticNbFiles()),
    Mount(f"{prefix}/static", app=StaticFilesOptionalAuth(directory=server.solara_static)),
    Route("/{fullpath:path}", endpoint=root),
]

app = Starlette(routes=routes, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)

# Uncomment the lines below to test solara mouted under a subpath
# def myroot(request: Request):
#     return JSONResponse({"framework": "solara"})

# routes_test_sub = [Route("/", endpoint=myroot), Mount("/foo/", routes=routes)]
# app = Starlette(routes=routes_test_sub, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)



import datetime as dt
from typing import Callable, Dict, List, Optional, Tuple, Union, cast

import ipyvue
import reacton

import solara
import solara.lab
from solara.components.input import _use_input_type


def use_close_menu(el: reacton.core.Element, is_open: solara.Reactive[bool]):
    is_open_ref = solara.use_ref(is_open)
    is_open_ref.current = is_open

    def monitor_events():
        def close_menu(*ignore_args):
            is_open_ref.current.set(False)

        widget = cast(ipyvue.VueWidget, solara.get_widget(el))
        widget.on_event("keyup.enter", close_menu)
        widget.on_event("keydown.tab", close_menu)

        def cleanup():
            widget.on_event("keyup.enter", close_menu, remove=True)
            widget.on_event("keydown.tab", close_menu, remove=True)

        return cleanup

    solara.use_effect(monitor_events, [])


@solara.component
def InputDate(
    value: Union[solara.Reactive[Optional[dt.date]], Optional[dt.date]],
    on_value: Optional[Callable[[Optional[dt.date]], None]] = None,
    label: str = "Pick a date",
    children: List[solara.Element] = [],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable[[bool], None]] = None,
    optional: bool = False,
    date_format: str = "%Y/%m/%d",
    first_day_of_the_week: int = 1,
    style: Optional[Union[str, Dict[str, str]]] = None,
    classes: Optional[List[str]] = None,
):
    """
    Show a textfield, which when clicked, opens a datepicker. The input date should be of type `datetime.date`.

    ## Basic Example

    ```solara
    import solara
    import solara.lab
    import datetime as dt


    @solara.component
    def Page():
        date = solara.use_reactive(dt.date.today())

        solara.lab.InputDate(date)
        solara.Text(str(date.value))
    ```

    ## Arguments

    * value: Reactive variable of type `datetime.date`, or `None`. This date is selected the first time the component is rendered.
    * on_value: a callback function for when value changes. The callback function receives the new value as an argument.
    * label: Text used to label the text field that triggers the datepicker.
    * children: List of Elements to be rendered under the calendar. If empty, a close button is rendered.
    * open_value: Controls and communicates the state of the datepicker. If True, the datepicker is open. If False, the datepicker is closed.
    Intended to be used in conjunction with a custom set of controls to close the datepicker.
    * on_open_value: a callback function for when open_value changes. Also receives the new value as an argument.
    * optional: Determines whether to show an error when value is `None`. If `True`, no error is shown.
    * date_format: Sets the format of the date displayed in the text field. Defaults to `"%Y/%m/%d"`. For more information, see
    <a href="https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes" target="_blank">the Python documentation</a>.
    * first_day_of_the_week: Sets the first day of the week, as an `int` starting count from Sunday (`=0`). Defaults to `1`, which is Monday.
    * style: CSS style to apply to the text field. Either a string or a dictionary of CSS properties (i.e. `{"property": "value"}`).
    * classes: List of CSS classes to apply to the text field.
    """
    value_reactive = solara.use_reactive(value, on_value)  # type: ignore
    del value, on_value
    datepicker_is_open = solara.use_reactive(open_value, on_open_value)  # type: ignore
    del open_value, on_open_value

    def set_date_typed_cast(value: Optional[str]):
        if value:
            try:
                date_value = dt.datetime.strptime(value, date_format).date()
                return date_value
            except ValueError:
                raise ValueError(f"Date {value} does not match format {date_format.replace('%', '')}")
        elif optional:
            return None
        else:
            raise ValueError("Date cannot be empty")

    def date_to_str(date: Optional[dt.date]) -> str:
        if date is not None:
            return date.strftime(date_format)
        return ""

    def set_date_cast(new_value: Optional[str]):
        if new_value:
            date_value = dt.datetime.strptime(new_value, "%Y-%m-%d").date()
            datepicker_is_open.set(False)
            value_reactive.value = date_value

    def standard_strfy(date: Optional[dt.date]):
        if date is None:
            return None
        else:
            return date.strftime("%Y-%m-%d")

    date_standard_str = standard_strfy(value_reactive.value)

    style_flat = solara.util._flatten_style(style)

    internal_value, error_message, set_value_cast = _use_input_type(value_reactive, set_date_typed_cast, date_to_str)

    if error_message:
        label += f" ({error_message})"
    input = solara.v.TextField(
        label=label,
        v_model=internal_value,
        on_v_model=set_value_cast,
        append_icon="mdi-calendar",
        error=bool(error_message),
        style_="min-width: 290px;" + style_flat,
        class_=", ".join(classes) if classes else "",
    )

    use_close_menu(input, datepicker_is_open)

    with solara.lab.Menu(
        activator=input,
        close_on_content_click=False,
        open_value=datepicker_is_open,
        use_activator_width=False,
    ):
        with solara.v.DatePicker(
            v_model=date_standard_str,
            on_v_model=set_date_cast,
            first_day_of_week=first_day_of_the_week,
            style_="width: 100%;",
        ):
            if len(children) > 0:
                solara.display(*children)


@solara.component
def InputDateRange(
    value: Union[solara.Reactive[Tuple[Optional[dt.date], Optional[dt.date]]], Tuple[Optional[dt.date], Optional[dt.date]]],
    on_value: Optional[Callable[[Optional[Tuple[Optional[dt.date], Optional[dt.date]]]], None]] = None,
    label: str = "Select dates",
    children: List[solara.Element] = [],
    open_value: Union[solara.Reactive[bool], bool] = False,
    on_open_value: Optional[Callable[[bool], None]] = None,
    optional: bool = False,
    date_format: str = "%Y/%m/%d",
    first_day_of_the_week: int = 1,
    style: Optional[Union[str, Dict[str, str]]] = None,
    classes: Optional[List[str]] = None,
):
    """
    Show a textfield, which when clicked, opens a datepicker that allows users to select a range of dates by choosing a starting and ending date.
    The input dates should be stored in a reactive tuple of type `datetime.date`. The list should contain either precisely two elements.
    For an empty pre-selection of dates, pass a reactive empty list.

    ## Basic Example

    ```solara
    import solara
    import solara.lab
    import datetime as dt


    @solara.component
    def Page():
        dates = solara.use_reactive(tuple([dt.date.today(), dt.date.today() + dt.timedelta(days=1)]))

        solara.lab.InputDateRange(dates)
        solara.Text(str(dates.value))
    ```

    ## Arguments

    * value: Tuple with elements of type `datetime.date`. For an empty pre-selection of dates, pass an empty tuple.
    * on_value: a callback function for when value changes. The callback function receives the new value as an argument.
    * label: Text used to label the text field that triggers the datepicker.
    * children: List of Elements to be rendered under the calendar. If empty, a close button is rendered.
    * open_value: Controls and communicates the state of the datepicker. If True, the datepicker is open. If False, the datepicker is closed.
    Intended to be used in conjunction with a custom set of controls to close the datepicker.
    * on_open_value: a callback function for when open_value changes. Also receives the new value as an argument.
    * date_format: Sets the format of the date displayed in the text field. Defaults to `"%Y/%m/%d"`. For more information,
    * optional: Determines whether go show an error when value is `None`. If `True`, no error is shown.
    see <a href="https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes" target="_blank">the Python documentation</a>.
    * first_day_of_the_week: Sets the first day of the week, as an `int` starting count from Sunday (`=0`). Defaults to `1`, which is Monday.
    * style: CSS style to apply to the text field. Either a string or a dictionary of CSS properties (i.e. `{"property": "value"}`).
    * classes: List of CSS classes to apply to the text field.

    ## A More Advanced Example

    ```solara
    import solara
    import solara.lab
    import datetime as dt


    @solara.component
    def Page():
        date = solara.use_reactive(tuple([dt.date.today(), None]))
        range_is_open = solara.use_reactive(False)
        stay_length = solara.use_reactive(1)

        def set_end_date(value):
            if value and value[0]:
                value = value[0]
                second_date = value + dt.timedelta(days=stay_length.value)
                date.set([value, second_date])

        def book():
            # Do some stuff here
            range_is_open.set(False)

        solara.use_memo(lambda: set_end_date(date.value), [stay_length.value])

        with solara.Column(style="width: 400px;"):
            solara.IntSlider("Length of stay", stay_length, min=1, max=10)
            with solara.lab.InputDateRange(
                date,
                on_value=set_end_date,
                open_value=range_is_open,
            ):
                with solara.Row(justify="end", style="width: 100%;"):
                    solara.Button(
                        label="Book",
                        color="primary",
                        on_click=book,
                    )

    ```
    """
    value_reactive = solara.use_reactive(value, on_value)  # type: ignore
    del value, on_value
    date_standard_strings = [date.strftime("%Y-%m-%d") for date in value_reactive.value if date is not None]
    datepicker_is_open = solara.use_reactive(open_value, on_open_value)  # type: ignore
    del open_value, on_open_value
    style_flat = solara.util._flatten_style(style)

    def dates_to_string(dates: Tuple[Optional[dt.date], Optional[dt.date]]):
        string_dates = [date.strftime(date_format) if date is not None else "" for date in dates]
        if (len(dates) < 2 or dates[1] is None) and not optional:
            return string_dates[0], "Please select two dates"
        return " - ".join(string_dates), None

    def set_dates_cast(values):
        date_value = cast(
            Tuple[Optional[dt.date], Optional[dt.date]], tuple([dt.datetime.strptime(item, "%Y-%m-%d").date() if item is not None else None for item in values])
        )
        if len(date_value) > 1 and date_value[1] is not None:
            datepicker_is_open.set(False)
        value_reactive.value = date_value

    string_dates, error_message = dates_to_string(value_reactive.value)

    if error_message:
        label += f" ({error_message})"
    input = solara.v.TextField(
        label=label,
        v_model=string_dates,
        append_icon="mdi-calendar",
        error=bool(error_message),
        style_="min-width: 290px;" + style_flat,
        readonly=True,
        class_=", ".join(classes) if classes else "",
    )

    # We include closing on tab in case users want to skip the field with tab
    use_close_menu(input, datepicker_is_open)

    with solara.lab.Menu(
        activator=input,
        close_on_content_click=False,
        open_value=datepicker_is_open,
        use_activator_width=False,
    ):
        with solara.v.DatePicker(
            v_model=date_standard_strings,
            on_v_model=set_dates_cast,
            range=True,
            first_day_of_week=first_day_of_the_week,
            style_="width: 100%;",
        ):
            if len(children) > 0:
                for el in children:
                    solara.display(el)



import sys

import micropip
import micropip._micropip
import micropip.package
import solara_bootstrap


class IOLoop:
    pass


class ZMQStream:
    pass


class MessageTracker:
    pass


class Context:
    pass


class Socket:
    pass


class Message:
    pass


NOBLOCK = 1
REQ = None
DEALER = 2
SUB = 3
POLLOUT = 4

sugar = solara_bootstrap
socket = solara_bootstrap
asyncio = solara_bootstrap
core = solara_bootstrap
error = solara_bootstrap
ioloop = solara_bootstrap


class ProfileDir:
    pass


class StdinNotImplementedError(Exception):
    pass


class Queue:
    pass


class QueueEmpty:
    pass


class InteractiveShell:
    pass


def display(*args, **kwargs):
    raise NotImplementedError


def clear_output(*args, **kwargs):
    raise NotImplementedError


def get_ipython():
    return None


class Process:
    pass


async def main():
    # fake some packages we don't need
    micropip.install("numpy")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["debugpy"] = micropip.package.PackageMetadata(name="debugpy", version="1.6", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["notebook"] = micropip.package.PackageMetadata(name="notebook", version="6.5", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["pyzmq"] = micropip.package.PackageMetadata(name="pyzmq", version="23.0", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["psutil"] = micropip.package.PackageMetadata(name="psutil", version="22.3", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["tornado"] = micropip.package.PackageMetadata(name="tornado", version="6.2", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["argon2-cffi-bindings"] = micropip.package.PackageMetadata(
        name="argon2-cffi-bindings", version="22.3", source=""
    )
    micropip._micropip.PACKAGE_MANAGER.installed_packages["ipython"] = micropip.package.PackageMetadata(name="ipython", version="8.3", source="")
    micropip._micropip.PACKAGE_MANAGER.installed_packages["jupyter_client"] = micropip.package.PackageMetadata(name="jupyter_client", version="7.3", source="")

    sys.modules["zmq"] = solara_bootstrap
    sys.modules["zmq.asyncio"] = solara_bootstrap
    sys.modules["zmq.eventloop"] = solara_bootstrap
    sys.modules["zmq.eventloop.ioloop"] = solara_bootstrap
    sys.modules["zmq.eventloop.zmqstream"] = solara_bootstrap

    for fake in (
        "psutil IPython IPython.core IPython.core.interactiveshell IPython.core.getipython IPython.core.error "
        "IPython.core.profiledir IPython.display tornado tornado.ioloop tornado.queues".split()
    ):
        sys.modules[fake] = solara_bootstrap

    requirements = [
        "Pygments==2.13.0",
        "ipywidgets<8",
        "pydantic",
        "jinja2",
        "bqplot",
        "altair",
        "vega_datasets",
        "plotly",
        "ipycanvas",
    ]
    for dep in requirements:
        await micropip.install(dep, keep_going=True)
    await micropip.install("/wheels/solara-1.29.0-py2.py3-none-any.whl", keep_going=True)
    import solara

    el = solara.Warning("lala")
    import solara

    solara.render_fixed(el)
    return el



"""# Title
"""

from typing import Optional, cast

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    title, set_title = solara.use_state(cast(Optional[str], "Custom title!"))
    with solara.VBox() as main:
        solara.ToggleButtonsSingle(value=title, values=[None, "Custom title!", "Different custom title"], on_value=set_title)

        if title is not None:
            # if the title is not set in a child component, the parent's title will be used
            with solara.Head():
                # title should always occur inside a Head component
                solara.Title(title)
            solara.Info(f"Your browser tab title should say {title}", classes=["mt-4"])
        else:
            solara.Warning("If no title is set, the parent title is used.", classes=["mt-4"])

    return main


__doc__ += apidoc(solara.Title.f)  # type: ignore



import threading
from pathlib import Path

import playwright
import playwright.sync_api
import reacton.ipywidgets as w

import solara

HERE = Path(__file__).parent


def test_large_cookie(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        port = solara_server.port
        # more than 8kb of cookies
        for i in range(9):
            browser.contexts[0].add_cookies([{"name": f"a_{i}", "value": "a" * 1024, "url": f"http://localhost:{port}"}])
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Examples").first.click()


def test_docs_basics(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Examples").first.click()

        page_session.locator("text=Calculator").first.click()
        page_session.locator("text=+/-").wait_for()
        page_session.screenshot(path="tmp/screenshot_calculator.png")

        page_session.locator("text=Bqplot").first.click()
        page_session.locator("text=Exponent").wait_for()
        page_session.screenshot(path="tmp/screenshot_bqplot.png")

        page_session.locator("text=Scatter plot using Plotly").first.click()
        page_session.locator("text=plotly express").first.wait_for()
        page_session.screenshot(path="tmp/screenshot_plotly.png")

        page_session.locator("text=Plotly Image Annotator").first.click()
        page_session.locator("text=how to annotate images with").first.wait_for()


@solara.component
def ClickButton():
    count, set_count = solara.use_state(0)
    if not isinstance(count, int):
        print("oops, state issue?")  # noqa
        count = 0

    def on_click():
        set_count(count + 1)

    return solara.Button(f"Clicked: {count}", on_click=on_click)


@solara.component
def ThreadTest():
    label, set_label = solara.use_state("initial")
    use_thread, set_use_thread = solara.use_state(False)

    def from_thread():
        set_label("from thread")

    def start_thread():
        if use_thread:
            thread = threading.Thread(target=from_thread)
            thread.start()
            return thread.join

    solara.use_side_effect(start_thread, [use_thread])
    # we need to trigger creating a new widget, to make sure we
    # invoke a solara.server.app.get_current_context
    if label == "initial":
        return w.Button(description=label, on_click=lambda: set_use_thread(True))
    else:
        return w.Label(value=label)


def test_from_thread(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:ThreadTest"):
        page_session.goto(solara_server.base_url)
        el = page_session.locator(".jupyter-widgets")
        assert el.text_content() == "initial"
        page_session.wait_for_timeout(500)
        el.click()
        page_session.locator("text=from thread").wait_for()


def test_state(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:ClickButton"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Clicked: 0").click()
        page_session.locator("text=Clicked: 1").click()
        # refresh...
        page_session.goto(solara_server.base_url)
        # and state should NOT be restored
        page_session.locator("text=Clicked: 0").wait_for()


def test_from_thread_two_users(browser: playwright.sync_api.Browser, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:ThreadTest"):
        context1 = browser.new_context()
        page1 = context1.new_page()
        context2 = browser.new_context()
        page2 = context2.new_page()

        page1.goto(solara_server.base_url)

        el1 = page1.locator(".jupyter-widgets")
        assert el1.text_content() == "initial"

        page2.goto(solara_server.base_url)
        el2 = page2.locator(".jupyter-widgets")
        assert el2.text_content() == "initial"

        page1.wait_for_timeout(500)
        page1.wait_for_timeout(500)

        el1.click()
        page1.locator("text=from thread").wait_for()

        page2.wait_for_timeout(500)
        assert el2.text_content() == "initial"

        el2.click()
        page2.locator("text=from thread").wait_for()
        page1.close()
        page2.close()


@solara.component
def StaticPage():
    return w.Label(value="Hello world")


def test_run_in_iframe(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    page_session.context.clear_cookies()
    with extra_include_path(HERE), solara_app("server_test:StaticPage"):
        page_session.set_content(
            f"""
            <html>
            <body>
                <iframe name="main" src="{solara_server.base_url}"></iframe>
            </body>
            </html>
        """
        )

        iframe = page_session.frame("main")
        el = iframe.locator(".jupyter-widgets")
        assert el.text_content() == "Hello world"



import threading
import typing
from typing import Callable, Optional, cast

import traitlets
from ipyvue import Template
from ipyvuetify.extra import FileInput
from ipywidgets import widget_serialization
from typing_extensions import TypedDict

import solara
import solara.hooks as hooks


class FileInfo(TypedDict):
    name: str
    size: int
    file_obj: typing.BinaryIO
    data: Optional[bytes]


class FileDropZone(FileInput):
    # override to narrow traitlet of FileInput
    template = traitlets.Instance(Template).tag(sync=True, **widget_serialization)
    template_file = (__file__, "file_drop.vue")
    items = traitlets.List(default_value=[]).tag(sync=True)
    label = traitlets.Unicode().tag(sync=True)


@solara.component
def FileDrop(
    label="Drop file here",
    on_total_progress: Optional[Callable[[float], None]] = None,
    on_file: Optional[Callable[[FileInfo], None]] = None,
    lazy: bool = True,
):
    """Region a user can drop a file into for file uploading.

    If lazy=True, no file content will be loaded into memory,
    nor will any data be transferred by default.
    A file object is passed to the `on_file` callback, and data will be transferred
    when needed.

    If lazy=False, the file content will be loaded into memory and passed to the `on_file` callback via the `.data` attribute.

    The on_file callback takes the following argument type:
    ```python
    class FileInfo(typing.TypedDict):
        name: str  # file name
        size: int  # file size in bytes
        file_obj: typing.BinaryIO
        data: Optional[bytes]: bytes  # only present if lazy=False
    ```


    ## Arguments
     * `on_total_progress`: Will be called with the progress in % of the file upload.
     * `on_file`: Will be called with a `FileInfo` object, which contains the file `.name`, `.length` and a `.file_obj` object.
     * `lazy`: Whether to load the file content into memory or not. If `False`,
        the file content will be loaded into memory and passed to the `on_file` callback via the `.data` attribute.

    """
    file_info, set_file_info = solara.use_state(None)
    wired_files, set_wired_files = solara.use_state(cast(Optional[typing.List[FileInfo]], None))

    file_drop = FileDropZone.element(label=label, on_total_progress=on_total_progress, on_file_info=set_file_info)  # type: ignore

    def wire_files():
        if not file_info:
            return

        real = cast(FileDropZone, solara.get_widget(file_drop))

        # workaround for @observe being cleared
        real.version += 1
        real.reset_stats()

        set_wired_files(cast(typing.List[FileInfo], real.get_files()))

    solara.use_side_effect(wire_files, [file_info])

    def handle_file(cancel: threading.Event):
        if not wired_files:
            return
        if on_file:
            if not lazy:
                wired_files[0]["data"] = wired_files[0]["file_obj"].read()
            else:
                wired_files[0]["data"] = None
            on_file(wired_files[0])

    result: solara.Result = hooks.use_thread(handle_file, [wired_files])
    if result.error:
        raise result.error

    return file_drop



import solara


@solara.component
def Page():
    solara.Markdown(
        """
    Hi there 👋! All fullscreen apps are linked from [the examples page](/examples).
    """
    )


@solara.component
def Layout(children):
    route, routes = solara.use_route()
    return children[0]



import reacton.ipyvuetify as rv
import solara

from ..data import dfs


@solara.component
def DataCard(name):
    df = dfs[name].df
    with rv.Card(max_width="400px") as main:
        with solara.Link(f"/tabular/{name}"):
            rv.Img(height="250", src=dfs[name].image_url)
        rv.CardTitle(children=[dfs[name].title])
        with rv.CardText():
            solara.Markdown(f"*{len(df):,} rows*")
            with solara.Link(f"/tabular/{name}"):
                solara.Button("Open table view", text=True, icon_name="mdi-table")
    return main


@solara.component
def Overview():
    with solara.ColumnsResponsive(12) as main:
        with solara.Card("Datasets"):
            with solara.ColumnsResponsive(12, small=6, large=4):
                for name in dfs:
                    DataCard(name)
    return main



from typing import Callable, Dict, List, Optional, Union

import reacton.ipyvuetify as v

import solara


@solara.value_component(bool)
def Switch(
    *,
    label: str = None,
    value: Union[bool, solara.Reactive[bool]] = True,
    on_value: Callable[[bool], None] = None,
    disabled: bool = False,
    children: list = [],
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """A switch component provides users the ability to choose between two distinct values. But aesthetically different from a checkbox.

    Basic examples

    ```solara
    import solara

    show_message = solara.reactive(True)
    disable = solara.reactive(False)


    @solara.component
    def Page():
        with solara.Column():
            with solara.Row():
                solara.Switch(label="Hide Message", value=show_message, disabled=disable.value)
                solara.Switch(label="Disable Message Switch", value=disable)

            if show_message.value:
                solara.Markdown("## Use Switch to show/hide message")

    ```


    ## Arguments

     * `label`: The label to display next to the switch.
     * `value`: The current value of the switch (True or False).
     * `on_value`: A callback that is called when the switch is toggled.
     * `disabled`: If True, the switch is disabled and cannot be used.
     * `children`: A list of child elements to display on the switch.
     * `classes`: Additional CSS classes to apply.
     * `style`: CSS style to apply.

    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    if label:
        children = [label] + children

    return v.Switch(
        label=label,
        v_model=reactive_value.value,
        on_v_model=reactive_value.set,
        disabled=disabled,
        class_=solara.util._combine_classes(classes),
        style_=solara.util._flatten_style(style),
        children=children,
    )



"""# FileBrowser

"""
from pathlib import Path
from typing import Optional, cast

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    file, set_file = solara.use_state(cast(Optional[Path], None))
    path, set_path = solara.use_state(cast(Optional[Path], None))
    directory, set_directory = solara.use_state(Path("~").expanduser())

    with solara.VBox() as main:
        can_select = solara.ui_checkbox("Enable select")

        def reset_path():
            set_path(None)
            set_file(None)

        # reset path and file when can_select changes
        solara.use_memo(reset_path, [can_select])
        solara.FileBrowser(directory, on_directory_change=set_directory, on_path_select=set_path, on_file_open=set_file, can_select=can_select)
        solara.Info(f"You are in directory: {directory}")
        solara.Info(f"You selected path: {path}")
        solara.Info(f"You opened file: {file}")
    return main


__doc__ += apidoc(solara.FileBrowser.f)  # type: ignore



import unittest.mock

import solara


def test_component_vue_basic():
    @solara._component_vue("component_vue_test.vue")
    def ComponentVueTest(value: int, name: str = "World"):
        pass

    box, rc = solara.render(ComponentVueTest(value=1))
    widget = box.children[0]
    assert widget.value == 1
    assert widget.name == "World"
    rc.render(ComponentVueTest(value=2, name="Universe"))
    assert widget.value == 2
    assert widget.name == "Universe"


def test_component_vue_callback():
    mock = unittest.mock.Mock()

    @solara._component_vue("component_vue_test.vue")
    def ComponentVueTest(value: int, on_value=None):
        pass

    box, rc = solara.render(ComponentVueTest(value=1, on_value=mock))
    widget = box.children[0]
    assert widget.value == 1
    mock.assert_not_called()
    widget.value = 2
    mock.assert_called_once_with(2)
    widget.value = 3
    mock.assert_called_with(3)


def test_component_vue_event():
    mock = unittest.mock.Mock()

    @solara._component_vue("component_vue_test.vue")
    def ComponentVueTest(event_foo=None):
        pass

    box, rc = solara.render(ComponentVueTest(event_foo=mock), handle_error=False)
    widget = box.children[0]
    mock.assert_not_called()
    widget._handle_event(None, {"event": "foo", "data": 42}, None)
    mock.assert_called_once_with(42)
    widget._handle_event(None, {"event": "foo", "data": 42}, [b"bar"])
    mock.assert_called_with(42, [b"bar"])



import os
from typing import Optional

import solara.util

from .minisettings import BaseSettings, Field
from .util import get_solara_home

try:
    import dotenv
except ImportError:
    pass
else:
    dotenv.load_dotenv()


home = get_solara_home()
if not home.exists():
    try:
        home.mkdir(parents=True, exist_ok=True)
    except OSError:
        pass  # can fail in for instance docker when $HOME is not set/writable


class Cache(BaseSettings):
    type: str = Field("memory", env="SOLARA_CACHE", title="Type of cache, e.g. 'memory', 'disk', 'redis', or a multilevel cache, e.g. 'memory,disk'")
    disk_max_size: str = Field("10GB", title="Maximum size for'disk' cache , e.g. 10GB, 500MB")
    memory_max_size: str = Field("1GB", title="Maximum size for 'memory-size' cache, e.g. 10GB, 500MB")
    memory_max_items: int = Field(128, title="Maximum number of items for 'memory' cache")
    clear: bool = Field(False, title="Clear the cache on startup, only applies to disk and redis caches")
    path: Optional[str] = Field(
        os.path.join(home, "cache"), env="SOLARA_CACHE_PATH", title="Storage location for 'disk' cache. Defaults to `${SOLARA_HOME}/cache`"
    )

    class Config:
        env_prefix = "solara_cache_"
        case_sensitive = False
        env_file = ".env"


# in colab or vscode there is not solara cdn proxy available
_should_use_proxy = not (solara.util.is_running_in_colab() or solara.util.is_running_in_vscode() or solara.util.is_running_in_voila())


class Assets(BaseSettings):
    cdn: str = "https://cdn.jsdelivr.net/npm/"
    proxy: bool = _should_use_proxy

    class Config:
        env_prefix = "solara_assets_"
        case_sensitive = False
        env_file = ".env"


assets: Assets = Assets()
cache: Cache = Cache()



import solara
from solara.alias import rv


@solara.component
def Hero(title, sub_title, button_text):
    with rv.Html(tag="section", class_="hero text-center") as main:
        with rv.Container():
            with rv.Row():
                with rv.Col(md=8, offset_md=2, sm=10, offset_sm=1):
                    rv.Html(tag="h1", children=[title], class_="mb-4"),
                    solara.HTML(tag="div", unsafe_innerHTML=f"<h2>{sub_title}</h2>", class_="mb-4"),
                    with solara.Link("/docs/quickstart"):
                        solara.Button(label=button_text, elevation=0, large=True, class_="btn-size--xlarge solara-docs-button")
    return main



"""
# CardActions

"""
import solara
import solara.lab
from solara.website.utils import apidoc

from . import NoPage

title = "CardActions"

Page = NoPage


__doc__ += apidoc(solara.CardActions.f)  # type: ignore



import json
import logging
import os
import subprocess
import sys
import warnings
from pathlib import Path
from typing import Optional

import IPython.display
from IPython.core.interactiveshell import InteractiveShell
from IPython.display import display

import solara
from solara.util import get_solara_home

HERE = Path(__file__).parent
logger = logging.getLogger(__name__)

jupyter_checked_path = get_solara_home() / ".jupyter_checked"
solara_checked_path = get_solara_home() / ".solara_checked"
solara_version = solara.__version__


def _should_perform_check(path: Path):
    if path.exists():
        return False
    try:
        home = get_solara_home()
        if not home.exists():
            home.mkdir(parents=True, exist_ok=True)
        # try writing, if we cannot, we will not check
        if not path.exists():
            path.write_text("")
            path.unlink()
    except OSError:
        return False
    return True


def should_perform_jupyter_check():
    if "PYTEST_CURRENT_TEST" in os.environ:
        return False
    return _should_perform_check(jupyter_checked_path)


def should_perform_solara_check():
    if "PYTEST_CURRENT_TEST" in os.environ:
        return False
    import solara.server.settings

    if solara.server.settings.main.mode == "production":
        return False
    return _should_perform_check(solara_checked_path)


@solara.component
def JupyterCheck():
    # We do 2 calls home:
    #  * 1 from pure js (should always work)
    #  * 1 from a widget (might not work)
    # This way we can see how many installations actually fail
    # Note that we only do this once, we touch ~/.solara/.jupyter_checked
    # to avoid doing it multiple times
    def inject_pure_js_check():
        # do this as an effect, otherwise we will use the display
        # that gets dispatched to solara, which will not come through
        # if the widgets do not work
        IPython.display.display(
            IPython.display.Javascript(
                data=f"""
const prevIframe = document.getElementById("solara-jupyter-check");
if(prevIframe)
    prevIframe.remove();
const iframe = document.createElement('iframe')
iframe.setAttribute("src", "https://solara.dev/static/public/success.html?check=purejs&version={solara_version}");
iframe.style.width = "0px";
iframe.style.height = "0px";
iframe.style.display = "none";
iframe.id = "solara-jupyter-check";
document.body.appendChild(iframe);
            """
            )
        )

    # this should always get through, even if widgets do not work
    solara.use_effect(inject_pure_js_check, [])

    def flag_jupyter_checked():
        try:
            jupyter_checked_path.write_text("")
        except OSError:
            pass

    solara.use_effect(flag_jupyter_checked, [])
    # this iframe should only get through if the widget installation succeeded
    return solara.v.Html(
        tag="iframe",
        attributes={"src": f"https://solara.dev/static/public/success.html?check=widget&version={solara_version}", "width": "0px", "height": "0px"},
        style_="display: none;",
    )


@solara.component
def SolaraCheck():
    def flag_solara_checked():
        try:
            solara_checked_path.write_text("")
        except OSError:
            pass

    solara.use_effect(flag_solara_checked, [])
    return solara.v.Html(
        tag="iframe",
        attributes={
            "src": f"https://solara.dev/static/public/success.html?system=solara&check=widget&version={solara_version}",
            "width": "0px",
            "height": "0px",
        },
        style_="display: none;",
    )


def getcmdline(pid):
    # for linux
    if sys.platform == "linux":
        with open(f"/proc/{pid}/cmdline", "rb") as f:
            return f.read().split(b"\00")[0].decode("utf-8")
    elif sys.platform == "darwin":
        return subprocess.check_output(["ps", "-o", "command=", "-p", str(pid)]).split(b"\n")[0].split(b" ")[0].decode("utf-8")
    elif sys.platform == "win32":
        return subprocess.check_output(["wmic", "process", "get", "commandline", "/format:list"]).split(b"\n")[0].split(b" ")[0].decode("utf-8")
    else:
        raise ValueError(f"Unsupported platform: {sys.platform}")


def get_server_python_executable(silent: bool = False):
    servers = []
    try:
        from jupyter_server import serverapp

        servers += list(serverapp.list_running_servers())
    except ImportError:
        pass
    try:
        from notebook import notebookapp

        servers += list(notebookapp.list_running_servers())
    except ImportError:
        pass

    pythons = [getcmdline(server["pid"]) for server in servers]
    if len(pythons) == 0:
        python = sys.executable
        if not silent:
            warnings.warn("Could not find servers, we are assuming the server is running under Python executable: %s" % python)
    elif len(pythons) > 1:
        info = "\n\t".join(pythons)
        if sys.executable in pythons:
            python = sys.executable
        else:
            python = pythons[0]
            if not silent:
                warnings.warn("Found multiple find servers:\n%s\n" "We are assuming the server is running under Python executable: %s" % (info, python))
    else:
        python = pythons[0]
    return python


libraries_minimal = [
    {"python": "ipyvuetify", "classic": "jupyter-vuetify/extension", "lab": "jupyter-vuetify"},
    {"python": "ipyvue", "classic": "jupyter-vue/extension", "lab": "jupyter-vue"},
]

libraries_extra = [
    {"python": "bqplot", "classic": "bqplot/extension", "lab": "bqplot"},
    {"python": "ipyvolume", "classic": "ipyvolume/extension", "lab": "ipyvolume"},
    {"python": "ipywebrtc", "classic": "jupyter-webrtc", "lab": "jupyter-webrtc"},
    {"python": "ipyleaflet", "classic": "ipyleaflet/extension", "lab": "ipyleaflet"},
]


def check_jupyter(
    server_python: Optional[str] = None,
    silent: bool = False,
    libraries: list = libraries_minimal,
    libraries_extra: list = libraries_extra,
    force: bool = False,
    extra: bool = False,
):
    if solara._using_solara_server():
        # for the server we don't need to do this check
        return
    if not InteractiveShell.initialized():
        # also, in a normal python repr, we don't want to display anything
        return
    try:
        python_executable = server_python or get_server_python_executable(silent)
        if Path(python_executable).resolve() != Path(sys.executable).resolve() or force:
            libraries_json = json.dumps(libraries + (libraries_extra if extra else []))
            display(
                IPython.display.Javascript(
                    data="""
                    window.jupyter_python_executable = %r;
                    window.jupyter_widget_checks_silent = %s;
                    window.jupyter_widget_checks_libraries = %s;
                    """
                    % (python_executable, str(silent).lower(), libraries_json)
                )
            )
            display(IPython.display.HTML(filename=str(HERE / "checks.html")))
        else:
            if not silent:
                display(
                    IPython.display.HTML(
                        data="<div>Jupyter server is running under the same Python executable as your kernel, no need to check 👍."
                        "<br> <i>Run solara.check_jupyter(force=True) to force checking.</i></div>"
                    )
                )
    except Exception:
        logger.exception("Could not check jupyter-widgets extensions.")


check_jupyter(silent=True)



import contextlib
import dataclasses
import logging
import sys
import threading
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from operator import getitem
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Generic,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import react_ipywidgets as react
import reacton.core
from reacton.utils import equals

import solara
from solara import _using_solara_server

T = TypeVar("T")
TS = TypeVar("TS")
S = TypeVar("S")  # used for state
logger = logging.getLogger("solara.toestand")

_DEBUG = False


class ThreadLocal(threading.local):
    reactive_used: Optional[Set["ValueBase"]] = None


thread_local = ThreadLocal()


# these hooks should go into react-ipywidgets
def use_sync_external_store(subscribe: Callable[[Callable[[], None]], Callable[[], None]], get_snapshot: Callable[[], Any]):
    _, set_counter = react.use_state(0)

    def force_update():
        set_counter(lambda x: x + 1)

    state = get_snapshot()
    prev_state = react.use_ref(state)

    def update_state():
        prev_state.current = state

    react.use_effect(update_state)

    def on_store_change(_ignore_new_state=None):
        new_state = get_snapshot()
        if not equals(new_state, prev_state.current):
            prev_state.current = new_state
            force_update()

    react.use_effect(lambda: subscribe(on_store_change), [])
    return state


def use_sync_external_store_with_selector(subscribe, get_snapshot: Callable[[], Any], selector):
    return use_sync_external_store(subscribe, lambda: selector(get_snapshot()))


def merge_state(d1: S, **kwargs) -> S:
    if dataclasses.is_dataclass(d1):
        return dataclasses.replace(d1, **kwargs)  # type: ignore
    if "pydantic" in sys.modules and isinstance(d1, sys.modules["pydantic"].BaseModel):
        return type(d1)(**{**d1.dict(), **kwargs})  # type: ignore
    return cast(S, {**cast(dict, d1), **kwargs})


class ValueBase(Generic[T]):
    def __init__(self, merge: Callable = merge_state):
        self.merge = merge
        self.listeners: Dict[str, Set[Tuple[Callable[[T], None], Optional[ContextManager]]]] = defaultdict(set)
        self.listeners2: Dict[str, Set[Tuple[Callable[[T, T], None], Optional[ContextManager]]]] = defaultdict(set)

    @property
    def lock(self):
        raise NotImplementedError

    @property
    def value(self) -> T:
        return self.get()

    @value.setter
    def value(self, value: T):
        self.set(value)

    def set(self, value: T):
        raise NotImplementedError

    def peek(self) -> T:
        raise NotImplementedError

    def get(self) -> T:
        raise NotImplementedError

    def _get_scope_key(self):
        raise NotImplementedError

    def subscribe(self, listener: Callable[[T], None], scope: Optional[ContextManager] = None):
        scope_id = self._get_scope_key()
        self.listeners[scope_id].add((listener, scope))

        def cleanup():
            self.listeners[scope_id].remove((listener, scope))

        return cleanup

    def subscribe_change(self, listener: Callable[[T, T], None], scope: Optional[ContextManager] = None):
        scope_id = self._get_scope_key()
        self.listeners2[scope_id].add((listener, scope))

        def cleanup():
            self.listeners2[scope_id].remove((listener, scope))

        return cleanup

    def fire(self, new: T, old: T):
        logger.info("value change from %s to %s, will fire events", old, new)
        scope_id = self._get_scope_key()
        scopes = set()
        for listener, scope in self.listeners[scope_id].copy():
            if scope is not None:
                scopes.add(scope)
        for listener2, scope in self.listeners2[scope_id].copy():
            if scope is not None:
                scopes.add(scope)
        stack = contextlib.ExitStack()
        with contextlib.ExitStack() as stack:
            for scope in scopes:
                stack.enter_context(scope)
            for listener, scope in self.listeners[scope_id].copy():
                listener(new)
            for listener2, scope in self.listeners2[scope_id].copy():
                listener2(new, old)

    def update(self, _f=None, **kwargs):
        if _f is not None:
            assert not kwargs
            with self.lock:
                kwargs = _f(self.get())
        with self.lock:
            # important to have this part thread-safe
            new = self.merge(self.get(), **kwargs)
            self.set(new)

    def use_value(self) -> T:
        # .use with the default argument doesn't give good type inference
        return self.use()

    def use(self, selector: Callable[[T], TS] = lambda x: x) -> TS:  # type: ignore
        return selector(self.value)

    def use_state(self) -> Tuple[T, Callable[[T], None]]:
        setter = self.set
        value = self.use()  # type: ignore
        return value, setter

    @property
    def fields(self) -> T:
        # we lie about the return type, but in combination with
        # setter we can make type safe setters (see docs/tests)
        return cast(T, Fields(self))

    def setter(self, field: TS) -> Callable[[TS], None]:
        _field = cast(FieldBase, field)

        def setter(new_value: TS):
            _field.set(new_value)

        return cast(Callable[[TS], None], setter)


# the default store for now, stores in a global dict, or when in a solara
# context, in the solara user context


class KernelStore(ValueBase[S], ABC):
    _global_dict: Dict[str, S] = {}  # outside of solara context, this is used
    # we keep a counter per type, so the storage keys we generate are deterministic
    _type_counter: Dict[Any, int] = defaultdict(int)
    scope_lock = threading.RLock()

    def __init__(self, key=None):
        super().__init__()
        self.storage_key = key
        self._global_dict = {}
        # since a set can trigger events, which can trigger new updates, we need a recursive lock
        self._lock = threading.RLock()
        self.local = threading.local()

    @property
    def lock(self):
        return self._lock

    def _get_scope_key(self):
        scope_dict, scope_id = self._get_dict()
        return scope_id

    def _get_dict(self):
        scope_dict = self._global_dict
        scope_id = "global"
        if _using_solara_server():
            import solara.server.kernel_context

            try:
                context = solara.server.kernel_context.get_current_context()
            except RuntimeError:  # noqa
                pass  # do we need to be more strict?
            else:
                scope_dict = cast(Dict[str, S], context.user_dicts)
                scope_id = context.id
        return cast(Dict[str, S], scope_dict), scope_id

    def peek(self):
        return self.get()

    def get(self):
        scope_dict, scope_id = self._get_dict()
        if self.storage_key not in scope_dict:
            with self.scope_lock:
                if self.storage_key not in scope_dict:
                    # we assume immutable, so don't make a copy
                    scope_dict[self.storage_key] = self.initial_value()
        return scope_dict[self.storage_key]

    def clear(self):
        scope_dict, scope_id = self._get_dict()
        if self.storage_key in scope_dict:
            del scope_dict[self.storage_key]

    def set(self, value: S):
        scope_dict, scope_id = self._get_dict()
        old = self.get()
        if equals(old, value):
            return
        scope_dict[self.storage_key] = value

        if _DEBUG:
            import traceback

            traceback.print_stack(limit=17, file=sys.stdout)

            print("change old", old)  # noqa
            print("change new", value)  # noqa

        self.fire(value, old)

    @abstractmethod
    def initial_value(self) -> S:
        pass


class KernelStoreValue(KernelStore[S]):
    default_value: S

    def __init__(self, default_value: S, key=None):
        self.default_value = default_value
        cls = type(default_value)
        if key is None:
            with KernelStoreValue.scope_lock:
                index = self._type_counter[cls]
                self._type_counter[cls] += 1
            key = cls.__module__ + ":" + cls.__name__ + ":" + str(index)
        super().__init__(key=key)

    def initial_value(self) -> S:
        return self.default_value


class KernelStoreFactory(KernelStore[S]):
    def __init__(self, factory: Callable[[], S], key=None):
        self.factory = factory
        try:
            prefix = factory.__qualname__
        except Exception:
            prefix = repr(factory)
        if key is None:
            with KernelStore.scope_lock:
                index = self._type_counter[prefix]
                self._type_counter[prefix] += 1
            try:
                key = factory.__module__ + ":" + prefix + ":" + str(index)
            except Exception:
                key = prefix + ":" + str(index)
        super().__init__(key=key)

    def initial_value(self) -> S:
        return self.factory()


class Reactive(ValueBase[S]):
    _storage: ValueBase[S]

    def __init__(self, default_value: Union[S, ValueBase[S]], key=None):
        super().__init__()
        if not isinstance(default_value, ValueBase):
            self._storage = KernelStoreValue(default_value, key=key)
        else:
            self._storage = default_value
        self.__post__init__()
        self._name = None
        self._owner = None

    def __set_name__(self, owner, name):
        self._name = name
        self._owner = owner

    def __repr__(self):
        value = self.peek()
        if self._name:
            return f"<Reactive {self._owner.__name__}.{self._name} value={value!r} id={hex(id(self))}>"
        else:
            return f"<Reactive value={value!r} id={hex(id(self))}>"

    def __str__(self):
        if self._name:
            return f"{self._owner.__name__}.{self._name}={self.value!r}"
        else:
            return f"{self.value!r}"

    @property
    def lock(self):
        return self._storage.lock

    def __post__init__(self):
        pass

    def update(self, *args, **kwargs):
        self._storage.update(*args, **kwargs)

    def set(self, value: S):
        if value is self:
            raise ValueError("Can't set a reactive to itself")
        self._storage.set(value)

    def get(self, add_watch=None) -> S:
        if add_watch is not None:
            warnings.warn("add_watch is deprecated, use .peek()", DeprecationWarning)
        if thread_local.reactive_used is not None:
            thread_local.reactive_used.add(self)
        # peek to avoid parents also adding themselves to the reactive_used set
        return self._storage.peek()

    def peek(self) -> S:
        """Return the value without automatically subscribing to listeners."""
        return self._storage.peek()

    def subscribe(self, listener: Callable[[S], None], scope: Optional[ContextManager] = None):
        return self._storage.subscribe(listener, scope=scope)

    def subscribe_change(self, listener: Callable[[S, S], None], scope: Optional[ContextManager] = None):
        return self._storage.subscribe_change(listener, scope=scope)

    def computed(self, f: Callable[[S], T]) -> "Computed[T]":
        def func():
            return f(self.get())

        return Computed(func, key=f.__qualname__)


class Singleton(Reactive[S]):
    _storage: KernelStore[S]

    def __init__(self, factory: Callable[[], S], key=None):
        import solara.server.kernel_context

        super().__init__(KernelStoreFactory(factory, key=key))

        # reset on kernel restart (e.g. hot reload)
        def reset():
            def cleanup():
                self._storage.clear()

            return cleanup

        solara.server.kernel_context.on_kernel_start(reset)

    def __set__(self, obj, value):
        raise AttributeError("Can't set a singleton")


class Computed(Reactive[S]):
    _storage: KernelStore[S]

    def __init__(self, f: Callable[[], S], key=None):
        self.f = f

        def on_change(*ignore):
            with self._auto_subscriber.value:
                self.set(f())

        import functools

        self._auto_subscriber = Singleton(functools.wraps(AutoSubscribeContextManager)(lambda: AutoSubscribeContextManager(on_change)))

        @functools.wraps(f)
        def factory():
            v = self._auto_subscriber.value
            with v:
                return f()

        super().__init__(KernelStoreFactory(factory, key=key))

        # reset on kernel restart (e.g. hot reload)
        def reset():
            def cleanup():
                self._storage.clear()

            return cleanup

        solara.server.kernel_context.on_kernel_start(reset)

    def __repr__(self):
        value = super().__repr__()
        return "<Computed" + value[len("<Reactive") : -1]


@overload
def computed(
    f: None,
    *,
    key: Optional[str] = ...,
) -> Callable[[Callable[[], T]], Reactive[T]]:
    ...


@overload
def computed(
    f: Callable[[], T],
    *,
    key: Optional[str] = ...,
) -> Reactive[T]:
    ...


def computed(
    f: Union[None, Callable[[], T]],
    *,
    key: Optional[str] = None,
) -> Union[Callable[[Callable[[], T]], Reactive[T]], Reactive[T]]:
    """Creates a reactive variable that is set to the return value of the function.

    The value will be updated when any of the reactive variables used in the function
    change.

    ## Example

    ```solara
    import solara
    import solara.lab


    a = solara.reactive(1)
    b = solara.reactive(2)

    @solara.lab.computed
    def total():
        return a.value + b.value

    def reset():
        a.value = 1
        b.value = 2

    @solara.component
    def Page():
        print(a, b, total)
        solara.IntSlider("a", value=a)
        solara.IntSlider("b", value=b)
        solara.Text(f"a + b = {total.value}")
        solara.Button("reset", on_click=reset)
    ```

    """

    def wrapper(f: Callable[[], T]):
        return Computed(f, key=key)

    if f is None:
        return wrapper
    else:
        return wrapper(f)


class ReactiveField(Reactive[T]):
    def __init__(self, field: "FieldBase"):
        # super().__init__()  # type: ignore
        # We skip the Reactive constructor, because we do not need it, but we do
        # want to be an instanceof for use in use_reactive
        ValueBase.__init__(self)
        self._field = field
        field = field
        while not isinstance(field, ValueBase):
            field = field._parent
        self._root = field
        assert isinstance(self._root, ValueBase)

    def __str__(self):
        return str(self._field)

    def __repr__(self):
        return f"<Reactive field {self._field}>"

    @property
    def lock(self):
        return self._root.lock

    def subscribe(self, listener: Callable[[T], None], scope: Optional[ContextManager] = None):
        def on_change(new, old):
            try:
                new_value = self._field.get(new)
            except IndexError:
                return  # the current design choice to silently drop the update message
            except KeyError:
                return  # same
            old_value = self._field.get(old)
            if not equals(new_value, old_value):
                listener(new_value)

        return self._root.subscribe_change(on_change, scope=scope)

    def subscribe_change(self, listener: Callable[[T, T], None], scope: Optional[ContextManager] = None):
        def on_change(new, old):
            try:
                new_value = self._field.get(new)
            except IndexError:
                return  # see subscribe
            except KeyError:
                return  # see subscribe
            old_value = self._field.get(old)
            if not equals(new_value, old_value):
                listener(new_value, old_value)

        return self._root.subscribe_change(on_change, scope=scope)

    def get(self, add_watch=None) -> T:
        if add_watch is not None:
            warnings.warn("add_watch is deprecated, use .peek()", DeprecationWarning)
        if thread_local.reactive_used is not None:
            thread_local.reactive_used.add(self)
        # peek to avoid parents also adding themselves to the reactive_used set
        return self._field.peek()

    def peek(self) -> T:
        return self._field.peek()

    def set(self, value: T):
        self._field.set(value)

    def update(self, *args, **kwargs):
        ValueBase.update(cast(ValueBase, self), *args, **kwargs)


def Ref(field: T) -> Reactive[T]:
    _field = cast(FieldBase, field)
    return cast(Reactive[T], ReactiveField[T](_field))


class FieldBase:
    _parent: Any

    def __getattr__(self, key):
        if key in ["_parent", "set", "_lock"] or key.startswith("__"):
            return self.__dict__[key]
        return FieldAttr(self, key)

    def __getitem__(self, key):
        return FieldItem(self, key)

    def get(self, obj=None):
        raise NotImplementedError

    def set(self, value):
        raise NotImplementedError


class Fields(FieldBase):
    def __init__(self, state: ValueBase):
        self._parent = state
        self._lock = state.lock

    def get(self, obj=None):
        # we are at the root, so override the object
        # so we can get the 'old' value
        if obj is not None:
            return obj
        return self._parent.get()

    def peek(self, obj=None):
        # we are at the root, so override the object
        # so we can get the 'old' value
        if obj is not None:
            return obj
        return self._parent.peek()

    def set(self, value):
        self._parent.set(value)

    def __repr__(self):
        return repr(self._parent)


class FieldAttr(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getattr(obj, self.key)

    def peek(self, obj=None):
        obj = self._parent.peek(obj)
        return getattr(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.peek()
            if isinstance(self.key, str):
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)
            else:
                raise TypeError(f"Type of key {self.key!r} is not supported")

    def __str__(self):
        return f".{self.key}"

    def __repr__(self):
        return f"<Field {self._parent}{self}>"


class FieldItem(FieldBase):
    def __init__(self, parent, key: str):
        self._parent = parent
        self.key = key
        self._lock = parent._lock

    def get(self, obj=None):
        obj = self._parent.get(obj)
        return getitem(obj, self.key)

    def peek(self, obj=None):
        obj = self._parent.peek(obj)
        return getitem(obj, self.key)

    def set(self, value):
        with self._lock:
            parent_value = self._parent.peek()
            if isinstance(self.key, int) and isinstance(parent_value, (list, tuple)):
                parent_type = type(parent_value)
                parent_value = parent_value.copy()  # type: ignore
                parent_value[self.key] = value
                self._parent.set(parent_type(parent_value))
            else:
                parent_value = merge_state(parent_value, **{self.key: value})
                self._parent.set(parent_value)


class AutoSubscribeContextManagerBase:
    # a render loop might trigger a new render loop of a differtent render context
    # so we want to save, and restore the current reactive_used
    reactive_used: Optional[Set[ValueBase]] = None
    reactive_added_previous_run: Optional[Set[ValueBase]] = None
    subscribed: Dict[ValueBase, Callable]

    def __init__(self):
        self.subscribed = {}

    def update_subscribers(self, change_handler, scope=None):
        assert self.reactive_used is not None
        reactive_used = self.reactive_used
        # remove subfields for which we already listen to it's root reactive value
        reactive_used_subfields = {k for k in reactive_used if isinstance(k, ValueSubField)}
        reactive_used = reactive_used - reactive_used_subfields
        # only add subfield for which we don't listen to it's parent
        for reactive_used_subfield in reactive_used_subfields:
            if reactive_used_subfield._root not in reactive_used:
                reactive_used.add(reactive_used_subfield)
        added = reactive_used - (self.reactive_added_previous_run or set())

        removed = (self.reactive_added_previous_run or set()) - reactive_used

        for reactive in added:
            if reactive not in self.subscribed:
                unsubscribe = reactive.subscribe_change(change_handler, scope=scope)
                self.subscribed[reactive] = unsubscribe
        for reactive in removed:
            unsubscribe = self.subscribed[reactive]
            unsubscribe()
            del self.subscribed[reactive]
        self.reactive_added_previous_run = added

    def unsubscribe_all(self):
        for reactive in self.subscribed:
            unsubscribe = self.subscribed[reactive]
            unsubscribe()

    def __enter__(self):
        self.reactive_used_before = thread_local.reactive_used
        self.reactive_used = thread_local.reactive_used = set()
        assert thread_local.reactive_used is self.reactive_used, f"{hex(id(thread_local.reactive_used))} vs {hex(id(self.reactive_used))}"

    def __exit__(self, exc_type, exc_val, exc_tb):
        thread_local.reactive_used = self.reactive_used_before


class AutoSubscribeContextManagerReacton(AutoSubscribeContextManagerBase):
    def __init__(self, element: solara.Element):
        self.element = element
        super().__init__()

    def __enter__(self):
        _, set_counter = solara.use_state(0, key="auto_subscribe_force_update_counter")

        def force_update(new_value, old_value):
            # can we do just x+1 to collapse multiple updates into one?
            set_counter(lambda x: x + 1)

        super().__enter__()

        def update_subscribers():
            self.update_subscribers(force_update, scope=reacton.core.get_render_context(required=True))

        solara.use_effect(update_subscribers, None)

        def on_close():
            def cleanup():
                assert self.reactive_added_previous_run is not None
                self.unsubscribe_all()

            return cleanup

        solara.use_effect(on_close, [])


class AutoSubscribeContextManager(AutoSubscribeContextManagerBase):
    on_change: Callable[[], None]

    def __init__(self, on_change: Callable[[], None]):
        super().__init__()
        self.on_change = on_change

    def __enter__(self):
        super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        value = super().__exit__(exc_type, exc_val, exc_tb)
        self.update_subscribers(self.on_change)
        return value


# alias for compatibility
State = Reactive
ValueSubField = ReactiveField

auto_subscribe_context_manager = AutoSubscribeContextManagerReacton
reacton.core._component_context_manager_classes.append(auto_subscribe_context_manager)



"""# Matplotlib

"""

import numpy as np
from matplotlib.figure import Figure

import solara
from solara.website.utils import apidoc

x = np.linspace(0, 2, 100)


@solara.component
def Page():
    freq, set_freq = solara.use_state(2.0)
    phase, set_phase = solara.use_state(0.1)
    y = np.sin(x * freq + phase)

    fig = Figure()
    ax = fig.subplots()
    ax.plot(x, y)
    ax.set_ylim(-1.2, 1.2)

    with solara.VBox() as main:
        solara.FloatSlider("Frequency", value=freq, on_value=set_freq, min=0, max=10)
        solara.FloatSlider("Phase", value=phase, on_value=set_phase, min=0, max=np.pi, step=0.1)
        solara.FigureMatplotlib(fig, dependencies=[freq, phase])
    return main


__doc__ += apidoc(solara.FigureMatplotlib.f)  # type: ignore



import ipyvue
import traitlets

import solara


class SpinnerSolaraWidget(ipyvue.VueTemplate):
    template_file = (__file__, "spinner-solara.vue")

    size = traitlets.Unicode("64px").tag(sync=True)


@solara.component
def SpinnerSolara(size="64px"):
    """Spinner component with the Solara logo to indicate the app is busy.

    ### Basic example

    ```solara
    import solara

    @solara.component
    def Page():
        solara.SpinnerSolara(size="100px")
    ```

    ## Arguments
     * `size`: Size of the spinner.
    """
    return SpinnerSolaraWidget.element(size=size)



import ipywidgets as widgets

clicks = 0

print("View button: I get run at startup, and for every page request")  # noqa


def on_click(button):
    global clicks
    clicks += 1
    button.description = f"Viewed {clicks} times"


button = widgets.Button(description="Never viewed")
button.on_click(on_click)

page = button



"""# Link

"""
import solara
from solara.website.utils import apidoc

routes = [
    solara.Route(path="/"),
    solara.Route(path="kiwi"),
    solara.Route(path="banana"),
    solara.Route(path="apple"),
]


@solara.component
def Page():
    route_current, routes = solara.use_route()
    with solara.VBox() as main:
        solara.Info("Note the address bar in the browser. It should change to the path of the link.")
        with solara.HBox():
            for route in routes:
                with solara.Link(route):
                    current = route_current is route
                    solara.Button(f"Go to {route.path}", color="red" if current else None)
    return main


__doc__ += apidoc(solara.Link.f)  # type: ignore



import functools
import logging
import os
import pdb
import sys
import threading
import traceback
import warnings
from typing import Any, Dict, MutableMapping
from unittest import mock

import ipykernel.kernelbase
import IPython.display
import ipywidgets
import ipywidgets.widgets.widget_output
from IPython.core.interactiveshell import InteractiveShell

from . import app, kernel_context, reload, settings
from .utils import pdb_guard

logger = logging.getLogger("solara.server.patch")
try:
    from reacton.patch_display import patch as patch_display
except:  # noqa
    patch_display = None  # type: ignore

if patch_display is not None and sys.platform != "emscripten":
    patch_display()
ipywidget_version_major = int(ipywidgets.__version__.split(".")[0])
ipykernel_version_major = int(ipykernel.__version__.split(".")[0])


class FakeIPython:
    def __init__(self, context: kernel_context.VirtualKernelContext):
        self.context = context
        self.kernel = context.kernel
        self.display_pub = self.kernel.shell.display_pub
        # needed for the pyplot interface of matplotlib
        # (although we don't really support it)
        self.events = mock.MagicMock()

    def enable_gui(self, gui):
        logger.error("ignoring call to enable_gui(%s)", gui)

    def register_post_execute(self, callback):
        # mpl requires this
        pass

    def set_parent(self, *args):
        pass

    def showtraceback(self):
        if settings.main.use_pdb:
            logger.exception("Exception, will be handled by debugger")
            pdb.post_mortem()
        etype, value, tb = sys.exc_info()
        traceback_string = "".join(traceback.format_exception(etype, value, tb))
        logger.error("Uncaught exception: %s", traceback_string)
        msg = {
            "type": "exception",
            "traceback": traceback_string,
        }

        for socket in self.context.control_sockets:
            try:
                socket.send_json(msg)
            except:  # noqa
                # TODO: should we remove it from the list?
                pass

    def magic(self, *args):
        # proplot requires this
        pass

    def set_custom_exc(self, exc_tuple, handler):
        # make dask work
        pass


Kernel_instance_original = ipykernel.kernelbase.Kernel.instance.__func__  # type: ignore


def kernel_instance_dispatch(cls, *args, **kwargs):
    if kernel_context.has_current_context():
        context = kernel_context.get_current_context()
        return context.kernel
    else:
        return Kernel_instance_original(cls, *args, **kwargs)


Kernel_initialized_initial = ipykernel.kernelbase.Kernel.initialized.__func__  # type: ignore


def kernel_initialized_dispatch(cls):
    if app is None:  # python is shutting down, and the comm dtor wants to send a close message
        return False
    if kernel_context.has_current_context():
        return True
    else:
        return Kernel_initialized_initial(cls)


InteractiveShell_instance_initial = InteractiveShell.instance.__func__  # type: ignore


def interactive_shell_instance_dispatch(cls, *args, **kwargs):
    if kernel_context.has_current_context():
        context = kernel_context.get_current_context()
        return context.kernel.shell
    else:
        return InteractiveShell_instance_initial(cls, *args, **kwargs)


def display_solara(
    *objs,
    include=None,
    exclude=None,
    metadata=None,
    transient=None,
    display_id=None,
    raw=False,
    clear=False,
    **kwargs,
):
    print(*objs)  # noqa


# from IPython.core.interactiveshell import InteractiveShell

# if transient is None:
#     transient = {}
# if metadata is None:
#     metadata = {}
# from IPython.core.display_functions import _new_id

# if display_id:
#     if display_id is True:
#         display_id = _new_id()
#     transient["display_id"] = display_id
# if kwargs.get("update") and "display_id" not in transient:
#     raise TypeError("display_id required for update_display")
# if transient:
#     kwargs["transient"] = transient

# if not objs and display_id:
#     # if given no objects, but still a request for a display_id,
#     # we assume the user wants to insert an empty output that
#     # can be updated later
#     objs = [{}]
#     raw = True

# if not raw:
#     format = InteractiveShell.instance().display_formatter.format

# if clear:
#     clear_output(wait=True)

# for obj in objs:
#     if raw:
#         publish_display_data(data=obj, metadata=metadata, **kwargs)
#     else:
#         format_dict, md_dict = format(obj, include=include, exclude=exclude)
#         if not format_dict:
#             # nothing to display (e.g. _ipython_display_ took over)
#             continue
#         if metadata:
#             # kwarg-specified metadata gets precedence
#             _merge(md_dict, metadata)
#         publish_display_data(data=format_dict, metadata=md_dict, **kwargs)
# if display_id:
#     return DisplayHandle(display_id)


def get_ipython():
    if kernel_context.has_current_context():
        context = kernel_context.get_current_context()
        our_fake_ipython = FakeIPython(context)
        return our_fake_ipython
    else:
        return None


class context_dict(MutableMapping):
    def _get_context_dict(self) -> dict:
        raise NotImplementedError

    def __delitem__(self, key) -> None:
        self._get_context_dict().__delitem__(key)

    def __getitem__(self, key):
        return self._get_context_dict().__getitem__(key)

    def __iter__(self):
        return self._get_context_dict().__iter__()

    def __len__(self):
        return self._get_context_dict().__len__()

    def __setitem__(self, key, value):
        self._get_context_dict().__setitem__(key, value)

    # support OrderedDict API for matplotlib
    def move_to_end(self, key, last=True):
        assert last, "only last=True is supported"
        item = self.pop(key)
        self[key] = item

    # matplotlib assumes .values() returns a list
    def values(self):
        return list(self._get_context_dict().values())


class context_dict_widgets(context_dict):
    def _get_context_dict(self) -> dict:
        if kernel_context.has_current_context():
            context = kernel_context.get_current_context()
            return context.widgets
        else:
            return global_widgets_dict


class context_dict_templates(context_dict):
    def _get_context_dict(self) -> dict:
        if kernel_context.has_current_context():
            context = kernel_context.get_current_context()
            return context.templates
        else:
            return global_templates_dict


class context_dict_user(context_dict):
    def __init__(self, name, default_dict):
        self.name = name
        self.default_dict = default_dict

    def _get_context_dict(self) -> dict:
        if kernel_context.has_current_context():
            context = kernel_context.get_current_context()
            if self.name not in context.user_dicts:
                context.user_dicts[self.name] = {}
            return context.user_dicts[self.name]
        else:
            return self.default_dict


def auto_watch_get_template(get_template):
    """Wraps get_template and adds a file listener for automatic .vue file reloading"""

    def wrapper(abs_path):
        template = get_template(abs_path)
        reload.reloader.watcher.add_file(abs_path)
        return template

    return wrapper


class ThreadDebugInfo:
    lock = threading.Lock()
    created = 0
    running = 0
    stopped = 0


Thread__init__ = threading.Thread.__init__
Thread__bootstrap = threading.Thread._bootstrap  # type: ignore


def WidgetContextAwareThread__init__(self, *args, **kwargs):
    Thread__init__(self, *args, **kwargs)
    with ThreadDebugInfo.lock:
        ThreadDebugInfo.created += 1

    self.current_context = None
    try:
        self.current_context = kernel_context.get_current_context()
    except RuntimeError:
        logger.debug(f"No context for thread {self}")


def WidgetContextAwareThread__bootstrap(self):
    with ThreadDebugInfo.lock:
        ThreadDebugInfo.running += 1
    try:
        _WidgetContextAwareThread__bootstrap(self)
    finally:
        with ThreadDebugInfo.lock:
            ThreadDebugInfo.running -= 1
            ThreadDebugInfo.stopped += 1


def _WidgetContextAwareThread__bootstrap(self):
    if not hasattr(self, "current_context"):
        return Thread__bootstrap(self)
    if self.current_context:
        # we need to call this manually, because set_context_for_thread
        # uses this, and the original _bootstrap calls it too late for us
        self._set_ident()
        kernel_context.set_context_for_thread(self.current_context, self)
        shell = self.current_context.kernel.shell
        shell.display_pub.register_hook(shell.display_in_reacton_hook)
    try:
        with pdb_guard():
            Thread__bootstrap(self)
    finally:
        if self.current_context:
            shell.display_pub.unregister_hook(shell.display_in_reacton_hook)


_patched = False
global_widgets_dict = {}
global_templates_dict: Dict[Any, Any] = {}
widgets = context_dict_widgets()


def Output_enter(self):
    self._flush()

    def hook(msg):
        if msg["msg_type"] == "display_data":
            self.outputs += ({"output_type": "display_data", "data": msg["content"]["data"], "metadata": msg["content"]["metadata"]},)
            return None
        if msg["msg_type"] == "clear_output":
            self.outputs = ()
            return None
        return msg

    ip = get_ipython()
    if ip:
        ip.display_pub.register_hook(hook)


def Output_exit(self, exc_type, exc_value, traceback):
    ip = get_ipython()
    if ip:
        ip.display_pub._hooks.pop()


def patch_ipyreact():
    import ipyreact
    import ipyreact.module

    from . import esm

    ipyreact.module.define_module = esm.define_module
    ipyreact.module.get_module_names = esm.get_module_names
    # define_module is also exported top level
    ipyreact.define_module = esm.define_module

    # make this a no-op, we'll create the widget when needed
    ipyreact.importmap._update_import_map = lambda: None


def once(f):
    called = False
    return_value = None

    @functools.wraps(f)
    def wrapper():
        nonlocal called
        nonlocal return_value
        if called:
            return return_value
        called = True
        return_value = f()
        return return_value

    return wrapper


@once
def patch_matplotlib():
    import matplotlib
    import matplotlib._pylab_helpers

    prev = matplotlib._pylab_helpers.Gcf.figs
    matplotlib._pylab_helpers.Gcf.figs = context_dict_user("matplotlib.pylab.figure_managers", prev)  # type: ignore

    RcParamsOriginal = matplotlib.RcParams
    counter = 0
    lock = threading.Lock()

    class RcParamsScoped(context_dict, matplotlib.RcParams):
        _was_initialized = False
        _without_kernel_dict: Dict[Any, Any]

        def __init__(self, *args, **kwargs) -> None:
            self._init()
            RcParamsOriginal.__init__(self, *args, **kwargs)

        def _init(self):
            nonlocal counter
            with lock:
                counter += 1
            self._user_dict_name = f"matplotlib.rcParams:{counter}"
            # this creates a copy of the CPython side of the dict
            self._without_kernel_dict = dict(zip(dict.keys(self), dict.values(self)))
            self._was_initialized = True

        def _set(self, key, val):
            # in matplotlib this directly calls dict.__setitem__
            # which would not call context_dict.__setitem__
            self[key] = val

        def _get(self, key):
            # same as _get
            return self[key]

        def _get_context_dict(self) -> dict:
            if not self._was_initialized:
                # since we monkey patch the class after __init__ was called
                # we may have to do that later on
                self._init()
            if kernel_context.has_current_context():
                context = kernel_context.get_current_context()
                if self._user_dict_name not in context.user_dicts:
                    # copy over the global settings when needed
                    context.user_dicts[self._user_dict_name] = self._without_kernel_dict.copy()
                return context.user_dicts[self._user_dict_name]
            else:
                return self._without_kernel_dict

    matplotlib.RcParams = RcParamsScoped
    matplotlib.rcParams.__class__ = RcParamsScoped
    # we chose to monkeypatch the class, instead of re-assiging to reParams for 2 reasons:
    # 1. the RcParams object could be imported in different namespaces
    # 2. the rcParams has extra methods, which means we have to otherwise monkeypatch the context_dict

    def cleanup():
        matplotlib._pylab_helpers.Gcf.figs = prev
        matplotlib.RcParams = RcParamsOriginal
        matplotlib.rcParams.__class__ = RcParamsOriginal

    return cleanup


def patch_heavy_imports():
    # patches that we only want to do if a package is imported, because they may slow down startup
    if "matplotlib" in sys.modules:
        patch_matplotlib()


def patch():
    global _patched
    global global_widgets_dict
    if _patched:
        warnings.warn("patch() called twice")
        return
    _patched = True
    __builtins__["display"] = IPython.display.display

    try:
        import ipyreact

        del ipyreact
    except ModuleNotFoundError:
        pass
    else:
        patch_ipyreact()

    if "MPLBACKEND" not in os.environ:
        if ipykernel_version_major < 6:
            # changed in https://github.com/ipython/ipykernel/pull/591
            os.environ["MPLBACKEND"] = "module://ipykernel.pylab.backend_inline"
        else:
            os.environ["MPLBACKEND"] = "module://matplotlib_inline.backend_inline"

    # the ipyvue.Template module cannot be accessed like ipyvue.Template
    # because the import in ipvue overrides it
    template_mod = sys.modules["ipyvue.Template"]
    template_mod.template_registry = context_dict_templates()  # type: ignore
    template_mod.get_template = auto_watch_get_template(template_mod.get_template)  # type: ignore

    # this module also imports get_template
    template_mod_vue = sys.modules["ipyvue.VueTemplateWidget"]
    template_mod_vue.get_template = template_mod.get_template  # type: ignore

    component_mod_vue = sys.modules["ipyvue.VueComponentRegistry"]
    component_mod_vue.vue_component_registry = context_dict_user("vue_component_registry", component_mod_vue.vue_component_registry)  # type: ignore
    component_mod_vue.vue_component_files = context_dict_user("vue_component_files", component_mod_vue.vue_component_files)  # type: ignore

    if ipywidget_version_major < 8:
        global_widgets_dict = ipywidgets.widget.Widget.widgets
        ipywidgets.widget.Widget.widgets = widgets  # type: ignore
    else:
        if hasattr(ipywidgets.widgets.widget, "_instances"):  # since 8.0.3
            global_widgets_dict = ipywidgets.widgets.widget._instances
            ipywidgets.widgets.widget._instances = widgets  # type: ignore
        elif hasattr(ipywidgets.widget.Widget, "_instances"):
            global_widgets_dict = ipywidgets.widget.Widget._instances
            ipywidgets.widget.Widget._instances = widgets  # type: ignore
        else:
            raise RuntimeError("Could not find _instances on ipywidgets version %r" % ipywidgets.__version__)
    threading.Thread.__init__ = WidgetContextAwareThread__init__  # type: ignore
    threading.Thread._bootstrap = WidgetContextAwareThread__bootstrap  # type: ignore
    # on CI we get a mypy error:
    # solara/server/patch.py:210: error: Cannot assign to a method
    #  solara/server/patch.py:210: error: Incompatible types in assignment (expression has type "classmethod[Any]",\
    #                                     variable has type "Callable[[VarArg(Any), KwArg(Any)], Any]")
    # not sure why we cannot reproduce that locally
    ipykernel.kernelbase.Kernel.instance = classmethod(kernel_instance_dispatch)  # type: ignore
    InteractiveShell.instance = classmethod(interactive_shell_instance_dispatch)  # type: ignore
    # on CI we get a mypy error:
    # solara/server/patch.py:211: error: Cannot assign to a method
    # solara/server/patch.py:211: error: Incompatible types in assignment (expression has type "classmethod[Any]", variable has type "Callable[[], Any]")
    # not sure why we cannot reproduce that locally
    ipykernel.kernelbase.Kernel.initialized = classmethod(kernel_initialized_dispatch)  # type: ignore
    ipywidgets.widgets.widget.get_ipython = get_ipython
    # TODO: find a way to actually monkeypatch get_ipython
    IPython.get_ipython = get_ipython

    ipywidgets.widgets.widget_output.Output.__enter__ = Output_enter
    ipywidgets.widgets.widget_output.Output.__exit__ = Output_exit

    original_close = ipywidgets.widget.Widget.close
    closed_ids = set()
    closed_stack: Dict[int, str] = {}

    def model_id_debug(self: ipywidgets.widgets.widget.Widget):
        from ipyvue.ForceLoad import force_load_instance

        import solara.comm

        if self.comm is None and id(self) in closed_ids and id(self) in closed_stack:
            raise RuntimeError(f"Widget {type(self)} has been closed, the stacktrace when the widget was closed is:\n{closed_stack[id(self)]}")

        if self.comm is None or isinstance(self.comm, solara.comm.DummyComm) and force_load_instance.comm is not self.comm:
            stack = solara.comm.orphan_comm_stacks.get(self.comm)
            if stack:
                raise RuntimeError(
                    "Widget has no comm, you are probably using a widget that was created at app startup, the stacktrace when the widget was created is:\n"
                    + stack
                )
            else:
                raise RuntimeError("Widget has no comm, you are probably using a widget that was closed. The widget is:\n" + repr(self))

        return self.comm.comm_id

    ipywidgets.widget.Widget.model_id = property(model_id_debug)

    def close_widget_debug(self: ipywidgets.widgets.widget.Widget):
        # only in development mode, since this leaks memory
        # can be called during shutdown/gc, so we need to check if the module is still there
        if settings and settings.main.mode == "development":
            stacktrace = "".join(traceback.format_stack())
            closed_stack[id(self)] = stacktrace
            closed_ids.add(id(self))
        original_close(self)

    ipywidgets.widget.Widget.close = close_widget_debug
